"""Classes defining config structure."""
from dataclasses import dataclass
from enum import Enum, unique


@unique
class ConfigTypes(Enum):
    """Enum class containing the valid types for config files."""

    GRIDDED = "gridded"
    INDEXED = "indexed"


class ConfigKeys:
    """Class of constants representing valid keys within configuriation json."""

    TYPE = "type"
    DIMENSIONALITY = "dimensionality"
    GRID_REF = "grid_ref"
    PROC_FLAGS = "processing_flags"
    DATASET = "dataset"
    DOMAIN = "domain"
    CODE_PROCESSING = "static_variables"
    DIM_MAP = "dimension_map"
    VAR_MAP = "variable_map"
    CHUNKS = "chunks"
    NO_GR_VAR = "not_grid_vars"
    COO_VAR = "coord_vars"
    DEL_VAR = "delete_vars"
    KEEP_ALL_VARS = "keep_all_vars"


@dataclass(frozen=True)
class DataFile:
    """General parent dataclass for holding common config attributes of datafiles.

    Args:
        variable_map (dict): dict containing mapping for variable names.
        dimension_map (dict): dict containing mapping for dimension names.
        keep_all_vars (boolean): True if xarray is to retain all data file variables
                                  otherwise False i.e keep only those in the json config file variable mappings.
    """

    variable_map: dict
    dimension_map: dict
    keep_all_vars: bool = False


@dataclass(frozen=True)
class CodeProcessing:
    """Dataclass holding config attributes for static variables that might not need changing between model runs

    Args:
        not_grid_variables (list): A list of variables not belonging to the grid.
        delete_variables (list):  A list of variables to drop from the dataset.
    """

    not_grid_variables: list
    delete_variables: list


@dataclass(frozen=True)
class Dataset(DataFile):
    """Dataclass holding config attributes for Dataset datafiles. Extends DataFile.

    Args:
        coord_var (list): list of dataset coordinate variables to apply once dataset is loaded
    """

    coord_var: list = None
    pass


@dataclass(frozen=True)
class Domain(DataFile):
    """Dataclass holding config attributes for Domain datafiles. Extends DataFile."""

    pass


@dataclass(frozen=True)
class Config:
    """General dataclass for holding common config file attributes.

    Args:
        dataset (Dataset): Dataset object representing 'dataset' config.
        processing_flags (list): List of processing flags.
        chunks (dict): dict for dask chunking config. (i.e. {"dim1":100, "dim2":100, "dim3":100}).
        type (ConfigTypes): Type of config. Must be a valid ConfigType.
    """

    dimensionality: int
    dataset: Dataset
    processing_flags: list
    chunks: dict
    type: ConfigTypes


@dataclass(frozen=True)
class GriddedConfig(Config):
    """Dataclass for holding gridded-config specific attributes. Extends Config.

    Args:
        type (ConfigTypes): Type of config. Set to ConfigTypes.GRIDDED.
        grid_ref (dict): dict containing key:value of grid_ref:[list of grid variables].
        domain (Domain): Domain object representing 'domain' config.
    """

    type: ConfigTypes = ConfigTypes.GRIDDED
    grid_ref: dict = None
    domain: Domain = None
    code_processing: CodeProcessing = None


@dataclass(frozen=True)
class IndexedConfig(Config):
    """Dataclass for holding indexed-config specific attributes. Extends Config.

    Args:
        type (ConfigTypes): Type of config. Set to ConfigTypes.INDEXED.
    """

    type: ConfigTypes = ConfigTypes.INDEXED



"""A class to convert from coast gridded to xesmf."""
import os.path as path_lib
import warnings
from ..data.gridded import Gridded

# from dask import delayed, compute, visualize
# import graphviz
import numpy as np
import xarray as xr


class xesmf_convert:
    """
    Converts the main dataset within a COAsT.Gridded object to be suitable
    for input to XESMF for regridding to either a curvilinear or rectilienar
    grid. All you need to do if provide a Gridded object and a grid type when
    creating a new instance of this class. It will then contain an appropriate
    input dataset. You may also provide a second COAsT gridded object if
    regridding between two objects. For using xesmf, please see the package's
    documentation website here:

    https://xesmf.readthedocs.io/en/latest/index.html

    You can install XESMF using:

        conda install -c conda-forge xesmf.

    The setup used by this class has been tested for xesmf v0.6.2 alongside
    esmpy v8.0.0. It was installed using:

        conda install -c conda-forge xesmf esmpy=8.0.0

    INPUTS
     input_gridded_obj (Gridded) :: Gridded object to be regridded
     output_gridded_obj(Gridded) :: (optional) Gridded object to regrid TO
     reorder_dims      (bool)    :: Xesmf requires that lat/lon dimensions are
                                    the last dimensions. If this is True, then
                                    will attempt to reorder dimensions. Not
                                    recommended for large datasets.
                                    [Default = False]

    >>> EXAMPLE USEAGE <<<
    If regridding a Gridded object to an arbitrarily defined rectilinear
    or curvilinear grid, you just need to do the following:

        import xesmf as xe

        # Create your gridded object
        gridded = coast.Gridded(*args, **kwargs)

        # Pass the gridded object over to xesmf_convert
        xesmf_ready = coast.xesmf_convert(gridded, input_grid_type = 'curvilinear')

        # Now this object will contain a dataset called xesmf_input, which can
        # be passed over to xesmf. E.G:

        destination_grid = xesmf.util.grid_2d(-15, 15, 1, 45, 65, 1)
        regridder = xe.Regridder(xesmf_ready.input_grid, destination_grid,
                                 "bilinear")
        regridded_dataset = regridder(xesmf_ready.input_data)

    XESMF contains a couple of difference functions for quickly creating output
    grids, such as xesmf.util.grid_2d and xesmf.util.grid_global(). See their
    website for more info.

    The process is almost the same if regridding from one COAsT.Gridded object
    to another (gridded0 -> gridded1):

        xesmf_ready = coast.xesmf_convert(gridded0, gridded1,
                                          input_grid_type = "curvilinear",
                                          output_grid_type = "curvilinear")
        regridder = xe.Regridder(xesmf_ready.input_grid,
                                 xesmf_ready.output_grid, "bilinear")
        regridded_dataset = regridder(xesmf_ready.input_data)

    Note that you can select which variables you want to regrid, either prior
    to using this tool or by indexing the input_data dataset. e.g.:

        regridded_dataset = regridder(xesmf_ready.input_data['temperature'])

    If your input datasets were lazy loaded, then so will the regridded dataset.
    At this point you can either load the data or (recomended) save the regridded
    data to file:

        regridded_dataset.to_netcdf(<filename_to_save>)

    Before saving back to file, call xesmf_ready.to_gridded() to convert
    the regridded xesmf object back to a gridded object

    """

    def __init__(
        self,
        input_gridded_obj=None,
        output_gridded_obj=None,
        input_grid_type="curvilinear",
        output_grid_type="curvilinear",
        reorder_dims=False,
    ):
        """
        Handles the inputs and reformatting of data. See Class docstring for
        more info.
        """

        # Set grid types
        self.input_grid_type = input_grid_type
        self.output_grid_type = output_grid_type

        # If an input object is provided, convert to grid/data for xesmf
        if input_gridded_obj is not None:
            input_dataset = input_gridded_obj.dataset
            self.input_grid, self.input_data = self._get_xesmf_datasets(input_dataset, input_grid_type)

        # If an output object is provided, convert to grid/data for xesmf
        if output_gridded_obj is not None:
            output_dataset = output_gridded_obj.dataset
            self.output_grid, _ = self._get_xesmf_datasets(output_dataset, output_grid_type)

    @classmethod
    def _get_xesmf_datasets(cls, dataset, grid_type, reorder_dims=False):
        """For a given dataset taken from a Gridded object and a grid_type
        (curvilinear or rectilinear), determine the xesmf formatted dataset.
        This method does some checks to make sure the dataset is suitable and
        renames the relevant dimensions/coordinates. Any vars that don't have
        both x_dim and y_dim will be dropped. If x_dim and y_dim are present
        BUT they are not the last dimensions AND reorder_dims=True then
        the dimensions will be reordered (not good for lazy loading/chunking).
        """

        # Check that x_dim and y_dim are in the dataset. If not, raise error
        if "x_dim" not in dataset.dims or "y_dim" not in dataset.dims:
            raise AttributeError("Gridded.dataset must  contain two spatial dimensions called x_dim and y_dim")

        # Check each variable has x_dim AND y_dim and that they are the
        # last dimensions. If not, then drop the variables if reorder_dims=False.
        # Otherwise, attempt to reorder dimensions.

        # Rename Dimensions and coordinates
        if grid_type == "curvilinear":
            dataset = dataset.swap_dims({"x_dim": "x", "y_dim": "y"})
            dataset = dataset.rename_vars({"latitude": "lat", "longitude": "lon"})

        # Define grid and data
        grid = dataset[["lon", "lat"]]
        data = dataset
        return grid, data

    @staticmethod
    def to_gridded(xesmf_dataset):
        """Converts an xesmf_dataset back to a Coast.Gridded() object. Returns
        a Gridded object."""

        # Create empty gridded object
        gridded_obj = Gridded()

        # Rename dimensions and coordinates
        xesmf_dataset = xesmf_dataset.swap_dims({"x": "x_dim", "y": "y_dim"})
        xesmf_dataset = xesmf_dataset.rename_vars({"lat": "latitude", "lon": "longitude"})

        # Place renamed dataset into new gridded object.
        gridded_obj.dataset = xesmf_dataset

        return gridded_obj



"""An analysis class for tide gauge."""
import numpy as np
import xarray as xr
from ..data.tidegauge import Tidegauge
from .._utils import general_utils, stats_util, crps_util
import matplotlib.dates as mdates
import utide as ut
import scipy.signal as signal


class TidegaugeAnalysis:
    """
    This contains analysis methods suitable for use with the dataset structure
    of Tidegauge()
    """

    def __init__(self):
        return

    @classmethod
    def match_missing_values(cls, data_array1, data_array2, fill_value=np.nan):
        """
        Will match any missing values between two tidegauge_multiple datasets.
        Where missing values (defined by fill_value) are found in either dataset
        they are also placed in the corresponding location in the other dataset.
        Returns two new tidegaugeMultiple objects containing the new
        ssh data. Datasets must contain ssh variables and only ssh will be
        masked.
        """

        if data_array2.dims[0] == "t_dim":
            data_array2 = data_array2.transpose()
        if data_array1.dims[0] == "t_dim":
            data_array1 = data_array1.transpose()

        if np.isnan(fill_value):
            ind1 = np.isnan(data_array1.values)
            ind2 = np.isnan(data_array2.values)
        else:
            ind1 = data_array1.values == fill_value
            ind2 = data_array2.values == fill_value

        ds1 = data_array1.where(~ind2)
        ds2 = data_array2.where(~ind1)

        return Tidegauge(dataset=ds1.to_dataset()), Tidegauge(dataset=ds2.to_dataset())

    @classmethod
    def harmonic_analysis_utide(
        cls,
        data_array,
        min_datapoints=1000,
        nodal=False,
        trend=False,
        method="ols",
        conf_int="linear",
        Rayleigh_min=0.95,
    ):
        """
        Does a harmonic analysis for each timeseries inside this object using
        the utide library. All arguments except min_datapoints are arguments
        that are passed to ut.solve(). Please see the utide  website for more
        information:

            https://pypi.org/project/UTide/

        Utide will by default do it's harmonic analysis using a set of harmonics
        determined using the Rayleigh criterion. This changes the number of
        harmonics depending on the length and frequency of the time series.

        Output from this routine is not a new dataset, but a list of utide
        analysis object. These are structures containing, amongst other things,
        amplitudes, phases, constituent names and confidence intervals. This
        list can be passed to reconstruct_tide_utide() in this object to create
        a new TidegaugeMultiple object containing reconstructed tide data.

        INPUTS
         data_array     : Xarray data_array from a coast.Tidegauge() object
                          e.g. tidegauge.dataset.ssh
         min_datapoints : If a time series has less than this value number of
                          datapoints, then omit from the analysis.
         <all_others>   : Inputs to utide.solve(). See website above.

        OUTPUTS
         A list of utide structures from the solve() routine. If a location
         is omitted, it will contain [] for it's entry.
        """
        # Make name shorter for computations and get dimension lengths
        ds = data_array
        n_port = ds.sizes["id_dim"]
        n_time = ds.sizes["t_dim"]

        # Harmonic analysis datenums -- for utide to work correctly < 0.3.0
        # time = mdates.date2num(ds.time.values)
        time = ds.time.values  # accepts np.datetime64 uTide >= 0.3.0

        # Create empty list of analyses
        analyses = []

        # Loop over ports
        for pp in range(0, n_port):

            # Temporary in-loop datasets
            ds_port = ds.isel(id_dim=pp).load()
            number_of_nan = np.sum(np.isnan(ds_port.values))

            # If not enough datapoints for analysis then append an empty list
            if (n_time - number_of_nan) < min_datapoints:
                analyses.append([])
                continue

            # Do harmonic analysis using UTide
            uts_obs = ut.solve(
                time,
                ds_port.values,
                lat=ds_port.latitude.values,
                nodal=nodal,
                trend=trend,
                method=method,
                conf_int=conf_int,
                Rayleigh_min=Rayleigh_min,
            )

            analyses.append(uts_obs)

        return analyses

    @classmethod
    def reconstruct_tide_utide(cls, data_array, utide_solution_list, constit=None, output_name="reconstructed"):
        """
        Use the tarray of times to reconstruct a time series series using a
        list of utide analysis objects. This list can be obtained
        using harmonic_analysis_utide(). Specify constituents to use in the
        reconstruction by passing a list of strings such as 'M2' to the constit
        argument. This won't work if a specified constituent is not present in
        the analysis.
        """

        # Get dimension lengths
        n_port = len(utide_solution_list)
        n_time = len(data_array.time)

        # Harmonic analysis datenums -- needed for utide < 0.3.0
        # time = mdates.date2num(data_array.time)
        time = data_array.time  # accepts np.datetime64 uTide >= 0.3.0

        # Get  coordinates from data_array and convert to Dataset for output
        reconstructed = np.zeros((n_port, n_time)) * np.nan

        # Loop over ports
        for pp in np.arange(n_port):

            # Reconstruct full tidal signal using utide
            pp_solution = utide_solution_list[pp]
            if len(pp_solution) == 0:
                continue

            # Call utide.reconstruct
            tide = np.array(ut.reconstruct(time, pp_solution, constit=constit).h)
            reconstructed[pp] = tide

        # Create output dataset and return it in new Tidegauge object.
        ds_out = xr.Dataset(data_array.coords)
        ds_out[output_name] = (["id_dim", "t_dim"], reconstructed)

        return Tidegauge(dataset=ds_out)

    @classmethod
    def calculate_non_tidal_residuals(
        cls, data_array_ssh, data_array_tide, apply_filter=True, window_length=25, polyorder=3
    ):
        """
        Calculate non tidal residuals by subtracting values in data_array_tide
        from data_array_ssh. You may optionally apply a filter to the non
        tidal residual data by setting apply_filter = True. This uses the
        scipy.signal.savgol_filter function, which you ay pass window_length
        and poly_order.
        """

        # NTR: Calculate non tidal residuals
        ntr = data_array_ssh - data_array_tide
        n_port = data_array_ssh.sizes["id_dim"]

        # NTR: Apply filter if wanted
        if apply_filter:
            for pp in range(n_port):
                ntr[pp, :] = signal.savgol_filter(ntr[pp, :], 25, 3)

        # Create output Tidegauge object and return
        ds_coords = data_array_ssh.coords.to_dataset()
        ds_coords["ntr"] = ntr
        return Tidegauge(dataset=ds_coords)

    @classmethod
    def threshold_statistics(cls, dataset, thresholds=np.arange(-0.4, 2, 0.1), peak_separation=12):
        """
        Do some threshold statistics for all variables with a time dimension
        inside this tidegauge_multiple object. Specifically, this routine will
        calculate:

                peak_count          : The number of indepedent peaks over
                                      each specified threshold. Independent peaks
                                      are defined using the peak_separation
                                      argument. This is the number of datapoints
                                      either side of a peak within which data
                                      is ommited for further peak search.
                time_over_threshold : The total time spent over each threshold
                                      This is NOT an integral, but simple a count
                                      of all points over threshold.
                dailymax_count      : A count of the number of daily maxima over
                                      each threshold
                monthlymax_count    : A count of the number of monthly maxima
                                      over each threshold.

        Output is a xarray dataset containing analysed variables. The name of
        each analysis variable is constructed using the original variable name
        and one of the above analysis categories.
        """

        # Set up working datasets and lists
        ds = dataset
        ds_thresh = xr.Dataset(ds.coords)
        ds_thresh["threshold"] = ("threshold", thresholds)
        var_list = list(ds.keys())
        n_thresholds = len(thresholds)
        n_port = ds.sizes["id_dim"]

        # Loop over vars in the input dataset
        for vv in var_list:

            empty_thresh = np.zeros((n_port, n_thresholds)) * np.nan
            ds_thresh["peak_count_" + vv] = (["id_dim", "threshold"], np.array(empty_thresh))
            ds_thresh["time_over_threshold_" + vv] = (["id_dim", "threshold"], np.array(empty_thresh))
            ds_thresh["dailymax_count_" + vv] = (["id_dim", "threshold"], np.array(empty_thresh))
            ds_thresh["monthlymax_count_" + vv] = (["id_dim", "threshold"], np.array(empty_thresh))

            for pp in range(n_port):

                # Identify NTR peaks for threshold analysis
                data_pp = ds[vv].isel(id_dim=pp)
                if np.sum(np.isnan(data_pp.values)) == ds.sizes["t_dim"]:
                    continue

                pk_ind, _ = signal.find_peaks(data_pp.values.copy(), distance=peak_separation)
                pk_values = data_pp[pk_ind]

                # Threshold Analysis
                for nn in range(0, n_thresholds):

                    # Calculate daily and monthly maxima for threshold analysis
                    ds_daily = data_pp.groupby("time.day")
                    ds_daily_max = ds_daily.max(skipna=True)
                    ds_monthly = data_pp.groupby("time.month")
                    ds_monthly_max = ds_monthly.max(skipna=True)

                    threshn = thresholds[nn]
                    # NTR: Threshold Frequency (Peaks)
                    ds_thresh["peak_count_" + vv][pp, nn] = np.sum(pk_values >= threshn)

                    # NTR: Threshold integral (Time over threshold)
                    ds_thresh["time_over_threshold_" + vv][pp, nn] = np.sum(data_pp >= threshn)

                    # NTR: Number of daily maxima over threshold
                    ds_thresh["dailymax_count_" + vv][pp, nn] = np.sum(ds_daily_max.values >= threshn)

                    # NTR: Number of monthly maxima over threshold
                    ds_thresh["monthlymax_count_" + vv][pp, nn] = np.sum(ds_monthly_max.values >= threshn)

        return ds_thresh

    @staticmethod
    def demean_timeseries(dataset):
        """
        Subtract time means from all variables within this tidegauge_multiple
        object. This is done independently for each id_dim location.
        """
        demeaned = dataset - dataset.mean(dim="t_dim")
        return Tidegauge(dataset=demeaned)

    @classmethod
    def difference(cls, dataset1, dataset2, absolute_diff=True, square_diff=True):
        """
        Calculates differences between two tide gauge objects datasets. Will calculate
        differences, absolute differences and square differences between all
        common variables within each object. Each object should have the same
        sized dimensions. When calling this routine, the differencing is done
        as follows:

            dataset1.difference(dataset2)

        This will do dataset1 - dataset2.

        Output is a new tidegauge object containing differenced variables.
        """

        # Get all differences and save coordintes for later
        differenced = dataset1 - dataset2
        diff_vars = list(differenced.keys())
        save_coords = list(dataset1.coords.keys())

        # Loop oer all variables
        for vv in diff_vars:
            differenced = differenced.rename({vv: "diff_" + vv})

        # Calculate absolute differences maybe
        if absolute_diff:
            abs_tmp = np.fabs(differenced)
            diff_vars = list(abs_tmp.keys())
            for vv in diff_vars:
                abs_tmp = abs_tmp.rename({vv: "abs_" + vv})
        else:
            abs_tmp = xr.Dataset()

        # Calculate squared differences maybe
        if square_diff:
            sq_tmp = np.square(differenced)
            diff_vars = list(sq_tmp.keys())
            for vv in diff_vars:
                sq_tmp = sq_tmp.rename({vv: "square_" + vv})
        else:
            sq_tmp = xr.Dataset()

        # Merge all differences into one
        differenced = xr.merge((differenced, abs_tmp, sq_tmp, dataset1[save_coords]))

        return Tidegauge(dataset=differenced)

    @staticmethod
    def find_high_and_low_water(data_array, method="comp", **kwargs):
        """
        Finds high and low water for a given variable.
        Returns in a new TIDEGAUGE object with similar data format to
        a TIDETABLE. If this Tidegauge object contains more than one location
        (id_dim > 1) then a list of Tidegauges will be returned.

        Methods:
        'comp' :: Find maxima by comparison with neighbouring values.
                  Uses scipy.signal.find_peaks. **kwargs passed to this routine
                  will be passed to scipy.signal.find_peaks.
        'cubic':: Find the maxima using the roots of cubic spline.
                  Uses scipy.interpolate.InterpolatedUnivariateSpline
                  and scipy.signal.argrelmax. **kwargs are not activated.
        NOTE: Currently only the 'comp' and 'cubic' methods implemented. Future
                  methods include linear interpolation or refinements.
        """

        if "id_dim" in data_array.dims:
            n_id = data_array.sizes["id_dim"]
        else:
            n_id = 1
            data_array = data_array.expand_dims("id_dim")

        tg_list = []
        # Loop over id_dim dimension
        for ii in range(n_id):
            # Get x and y for input to find_maxima

            x = data_array.isel(id_dim=ii).time.values
            y = data_array.isel(id_dim=ii).values

            # Get time and values of maxima and minima
            time_max, values_max = stats_util.find_maxima(x, y, method=method, **kwargs)
            time_min, values_min = stats_util.find_maxima(x, -y, method=method, **kwargs)
            # Place the above values into a brand new dataset for this id_dim index
            new_dataset = xr.Dataset()
            new_dataset.attrs = data_array.attrs
            new_dataset[data_array.name + "_highs"] = ("time_highs", values_max)
            new_dataset[data_array.name + "_lows"] = ("time_lows", -values_min)
            new_dataset["time_highs"] = ("time_highs", time_max)
            new_dataset["time_lows"] = ("time_lows", time_min)

            # Place dataset into a new Tidegauge object and append to output
            new_object = Tidegauge(dataset=new_dataset)
            tg_list.append(new_object)

        # If only 1 index, return just a Tidegauge object, else return list.
        if n_id == 1:
            return new_object
        else:
            return tg_list

    @staticmethod
    def doodson_x0_filter(dataset, var_str):
        """Applies doodson X0 filter to a specified TIDEGAUGE variable
        Input ius expected to be hourly. Use resample_mean to average data
        to hourly frequency."""
        n_id = dataset.sizes["id_dim"]
        n_time = dataset.sizes["t_dim"]
        filtered = np.zeros((n_id, n_time))
        for ii in range(n_id):
            ds_ii = dataset.isel(id_dim=ii)
            filtered[ii] = stats_util.doodson_x0_filter(ds_ii[var_str], ax=0)
        new_dataset = xr.Dataset(dataset.coords)
        new_dataset[var_str] = (("id_dim", "t_dim"), filtered)
        return Tidegauge(dataset=new_dataset)

    @classmethod
    def crps(
        cls,
        tidegauge_data,
        gridded_data,
        nh_radius: float = 20,
        time_interp: str = "linear",
    ):
        """
        Comparison of observed variable to modelled using the Continuous
        Ranked Probability Score. This is done using this TIDEGAUGE object.
        This method specifically performs a single-observation neighbourhood-
        forecast method.

        Parameters
        ----------
        model_object (model) : Model object (NEMO) containing model data
        model_var_name (str) : Name of model variable to compare.
        obs_var_name (str)   : Name of observed variable to compare.
        nh_radius (float)    : Neighbourhood rad
        cdf_type (str)       : Type of cumulative distribution to use for the
                               model data ('empirical' or 'theoretical').
                               Observations always use empirical.
        time_interp (str)    : Type of time interpolation to use (s)
        create_new_obj (bool): If True, save output to new TIDEGAUGE obj.
                               Otherwise, save to this obj.

        Returns
        -------
        xarray.Dataset containing times, sealevel and quality control flags

        Example Useage
        -------
        # Compare modelled 'sossheig' with 'ssh' using CRPS
        crps = altimetry.crps(nemo, 'sossheig', 'ssh')
        """

        # Extract variables and get shape
        mod_var = gridded_data
        obs_var = tidegauge_data
        n_id, n_time = obs_var.shape

        # Make output CRPS array
        crps_out = np.zeros((n_id, n_time))
        N_out = np.zeros((n_id, n_time))

        # Loop over location indices
        for ii in range(n_id):

            obs_ii = obs_var.isel(id_dim=ii)

            # Calculate CRPS
            crps_list, n_model_pts, contains_land = crps_util.crps_sonf_fixed(
                mod_var,
                tidegauge_data.longitude.values,
                tidegauge_data.latitude.values,
                obs_ii.values,
                tidegauge_data.time.values,
                nh_radius,
                time_interp,
            )

            crps_out[ii] = crps_list
            N_out[ii] = n_model_pts

        # Put into new object
        new_dataset = xr.Dataset(tidegauge_data.coords)
        new_dataset["crps"] = (("id_dim", "t_dim"), crps_out)
        new_dataset["crps_N"] = (("id_dim", "t_dim"), N_out)
        return Tidegauge(dataset=new_dataset)

    @classmethod
    def time_mean(cls, dataset, date0=None, date1=None):
        """Time mean of all variables between dates date0, date1"""
        var = general_utils.data_array_time_slice(dataset, date0, date1)
        return Tidegauge(dataset=var.mean(dim="t_dim", skipna=True))

    @classmethod
    def time_std(cls, dataset, date0=None, date1=None):
        """Time st. dev of variable var_str between dates date0 and date1"""
        var = general_utils.data_array_time_slice(dataset, date0, date1)
        return Tidegauge(dataset=var.std(dim="t_dim", skipna=True))

    @classmethod
    def time_slice(cls, dataset, date0=None, date1=None):
        sliced = general_utils.data_array_time_slice(dataset, date0, date1)
        return Tidegauge(dataset=sliced)

    @classmethod
    def resample_mean(cls, dataset, time_freq: str, **kwargs):
        """Resample a TIDEGAUGE variable in time by calculating the mean
            of all data points at a given frequency.

        Parameters
        ----------
        time_freq (str)  : Time frequency. e.g. '1H' for hourly, '1D' for daily
                           Can also be a timedelta object. See Pandas resample
                           method for more info.
        **kwargs (other) : Other arguments to pass to xarray.Dataset.resample
        (http://xarray.pydata.org/en/stable/generated/xarray.Dataset.resample.html)

        Returns
        -------
        New Tidegauge() object containing resampled data
        """

        # Resample using xarray.resample
        dataset = dataset.swap_dims({"t_dim": "time"})
        resampled = dataset.resample(time=time_freq, **kwargs).mean()
        resampled = resampled.swap_dims({"time": "t_dim"})
        return Tidegauge(dataset=resampled)



"""Module with attributes defining month ranges for the four seasons.

Used for convenience with Climatology.multiyear_averages().
Note: Summer is defined as JJAS, as opposed to the meteorological seasons of JJA.
"""

SPRING = [(3, 5)]
SUMMER = [(6, 9)]
AUTUMN = [(10, 11)]
WINTER = [(12, 2)]
ALL = [(3, 5), (6, 9), (10, 11), (12, 2)]



"""Mask maker"""
import xarray as xr
import numpy as np
import skimage.draw as draw
from . import general_utils


class MaskMaker:
    def __init__(self):

        return

    @staticmethod
    def make_mask_dataset(longitude, latitude, mask_list):
        if type(mask_list) is not list:
            mask_list = [mask_list]
        gridded_mask = xr.Dataset()
        gridded_mask["longitude"] = (["y_dim", "x_dim"], longitude)
        gridded_mask["latitude"] = (["y_dim", "x_dim"], latitude)
        n_masks = len(mask_list)
        nr, nc = mask_list[0].shape
        all_masks = np.zeros((n_masks, nr, nc))
        gridded_mask["mask"] = (["dim_mask", "y_dim", "x_dim"], all_masks)
        for mm in np.arange(n_masks):
            gridded_mask["mask"][mm] = mask_list[mm]
        return gridded_mask

    @staticmethod
    def fill_polygon_by_index(array_to_fill, vertices_r, vertices_c, fill_value=1, additive=False):
        """
        Draws and fills a polygon onto an existing numpy array based on array
        indices. To create a new mask, give np.zeros(shape) as input.
        Polygon vertices are drawn in the order given.

        Parameters
        ----------
        array_to_fill (2D array): Array onto which to fill polygon
        vertices_r (1D array): Row indices for polygon vertices
        vertices_c (1D_array): Column indices for polygon vertices
        fill_value (float, bool or int): Fill value for polygon (Default: 1)
        additive (bool): If true, add fill value to existing array. Otherwise
                         indices will be overwritten. (Default: False)

        Returns
        -------
        Filled 2D array
        """
        array_to_fill = np.array(array_to_fill)
        polygon_ind = draw.polygon(vertices_r, vertices_c, array_to_fill.shape)
        if additive:
            array_to_fill[polygon_ind[0], polygon_ind[1]] += fill_value
        else:
            array_to_fill[polygon_ind[0], polygon_ind[1]] = fill_value
        return array_to_fill

    @staticmethod
    def fill_polygon_by_lonlat(
        array_to_fill, longitude, latitude, vertices_lon, vertices_lat, fill_value=1, additive=False
    ):
        """
        Draws and fills a polygon onto an existing numpy array based on
        vertices defined by longitude and latitude locations. This does NOT
        draw a polygon on a sphere, but instead based on straight lines
        between points. This is OK for small regional areas, but not advisable
        for large and global regions.
        Polygon vertices are drawn in the order given.

        Parameters
        ----------
        array_to_fill (2D array): Array onto which to fill polygon
        vertices_r (1D array): Row indices for polygon vertices
        vertices_c (1D_array): Column indices for polygon vertices
        fill_value (float, bool or int): Fill value for polygon (Default: 1)
        additive (bool): If true, add fill value to existing array. Otherwise
                         indices will be overwritten. (Default: False)

        Returns
        -------
        Filled 2D array
        """
        array_to_fill = np.array(array_to_fill)
        ind_2d = general_utils.nearest_indices_2d(longitude, latitude, vertices_lon, vertices_lat)

        polygon_ind = draw.polygon(ind_2d[1], ind_2d[0], array_to_fill.shape)
        if additive:
            array_to_fill[polygon_ind[0], polygon_ind[1]] += fill_value
        else:
            array_to_fill[polygon_ind[0], polygon_ind[1]] = fill_value
        return array_to_fill

    @classmethod
    def region_def_nws_north_sea(cls, longitude, latitude, bath):
        """
        Regional definition for the North Sea (Northwest European Shelf)
        Longitude, latitude and bath should be 2D arrays corresponding to model
        coordinates and bathymetry. Bath should be positive with depth.
        """
        vertices_lon = [-5.34, -0.7, 7.5, 7.5, 9, 9, 6.3, 6.3, 5, 5, 4.126, 4.126, -1.071]
        vertices_lat = [56.93, 54.09, 54.09, 56, 56, 57.859, 57.859, 58.121, 58.121, 58.59, 58.59, 60.5, 60.5]

        mask = cls.fill_polygon_by_lonlat(np.zeros(longitude.shape), longitude, latitude, vertices_lon, vertices_lat)
        mask = mask * (bath < 200) * (bath > 0) * (~np.isnan(bath))
        return mask

    @classmethod
    def region_def_nws_outer_shelf(cls, longitude, latitude, bath):
        """
        Regional definition for the Outher Shelf (Northwest European Shelf)
        Longitude, latitude and bath should be 2D arrays corresponding to model
        coordinates and bathymetry. Bath should be positive with depth.
        """
        vertices_lon = [-4, -9.5, -1, 3.171, 3.171, -3.76, -3.76, -12, -12, -12, -4]
        vertices_lat = [50.5, 52.71, 60.5, 60.45, 63.3, 63.3, 60.45, 60.45, 55.28, 48, 48]
        mask = cls.fill_polygon_by_lonlat(np.zeros(longitude.shape), longitude, latitude, vertices_lon, vertices_lat)
        mask = mask * (bath < 200) * (bath > 0) * (~np.isnan(bath))
        return mask

    @classmethod
    def region_def_nws_norwegian_trench(cls, longitude, latitude, bath):
        """
        Regional definition for the Norwegian Trench (Northwest European Shelf)
        Longitude, latitude and bath should be 2D arrays corresponding to model
        coordinates and bathymetry. Bath should be positive with depth.
        """
        vertices_lon = [10.65, 1.12, 1.12, 10.65]
        vertices_lat = [61.83, 61.83, 48, 48]
        mask = cls.fill_polygon_by_lonlat(np.zeros(longitude.shape), longitude, latitude, vertices_lon, vertices_lat)
        mask = mask * (bath > 200) * (bath > 0) * (~np.isnan(bath))
        return mask

    @classmethod
    def region_def_nws_english_channel(cls, longitude, latitude, bath):
        """
        Regional definition for the English Channel (Northwest European Shelf)
        Longitude, latitude and bath should be 2D arrays corresponding to model
        coordinates and bathymetry. Bath should be positive with depth.
        """
        vertices_lon = [7.57, 7.57, -0.67, -2, -3.99, -3.99, -3.5, 12, 14]
        vertices_lat = [56, 54.08, 54.08, 50.7, 50.7, 48.8, 48, 48, 56]
        mask = cls.fill_polygon_by_lonlat(np.zeros(longitude.shape), longitude, latitude, vertices_lon, vertices_lat)
        mask = mask * (bath < 200) * (bath > 0) * (~np.isnan(bath))
        return mask

    @classmethod
    def region_def_south_north_sea(cls, longitude, latitude, bath):
        vertices_lon = [-0.67, -0.67, 9, 9, 7.57, 7.57]
        vertices_lat = [54.08, 51, 51, 56, 56, 54.08]
        mask = cls.fill_polygon_by_lonlat(np.zeros(longitude.shape), longitude, latitude, vertices_lon, vertices_lat)
        mask = mask * (bath < 200) * (bath > 0) * (~xr.ufuncs.isnan(bath))
        return mask

    @classmethod
    def region_def_off_shelf(cls, longitude, latitude, bath):
        vertices_lon = [10, 10, -5, -10, 0, 0, -20, -20]
        vertices_lat = [65, 60, 59, 52.5, 47.5, 45, 40, 63]
        mask = cls.fill_polygon_by_lonlat(np.zeros(longitude.shape), longitude, latitude, vertices_lon, vertices_lat)
        mask = mask * (bath > 200) * (bath > 0) * (~xr.ufuncs.isnan(bath))
        return mask

    @classmethod
    def region_def_irish_sea(cls, longitude, latitude, bath):
        vertices_lon = [-5, -7.6, -7.5, -4.1, 0, -2.6]
        vertices_lat = [56.4, 55, 52, 50.7, 51.5, 55.3]
        mask = cls.fill_polygon_by_lonlat(np.zeros(longitude.shape), longitude, latitude, vertices_lon, vertices_lat)
        mask = mask * (bath < 200) * (bath > 0) * (~xr.ufuncs.isnan(bath))
        return mask

    @classmethod
    def region_def_kattegat(cls, longitude, latitude, bath):
        vertices_lon = [9, 9, 13, 13]
        vertices_lat = [60, 52.5, 52.5, 60]
        mask = cls.fill_polygon_by_lonlat(np.zeros(longitude.shape), longitude, latitude, vertices_lon, vertices_lat)
        mask = mask * (bath < 200) * (bath > 0) * (~xr.ufuncs.isnan(bath))
        return mask



"""
Python definitions used to aid with statistical calculations.

*Methods Overview*
    -> normal_distribution(): Create values for a normal distribution
    -> cumulative_distribution(): Integration udner a PDF
    -> empirical_distribution(): Estimates CDF empirically
"""

import numpy as np
import xarray as xr
import scipy

from .logging_util import error


def quadratic_spline_roots(spl):
    """
    A custom function for the roots of a quadratic spline. Cleverness found at
    https://stackoverflow.com/questions/50371298/find-maximum-minimum-of-a-1d-interpolated-function
    Used in find_maxima().

    Example usage:
    see example_scripts/tidegauge_tutorial.py
    """
    roots = []
    knots = spl.get_knots()
    for a, b in zip(knots[:-1], knots[1:]):
        u, v, w = spl(a), spl((a + b) / 2), spl(b)
        t = np.roots([u + w - 2 * v, w - u, 2 * v])
        t = t[np.isreal(t) & (np.abs(t) <= 1)]
        roots.extend(t * (b - a) / 2 + (b + a) / 2)
    return np.sort(roots)


def find_maxima(x, y, method="comp", **kwargs):
    """
    Finds maxima of a time series y. Returns maximum values of y (e.g heights)
    and corresponding values of x (e.g. times).
    **kwargs are dependent on method.

        Methods:
        'comp' :: Find maxima by comparison with neighbouring values.
                  Uses scipy.signal.find_peaks. **kwargs passed to this routine
                  will be passed to scipy.signal.find_peaks.
        'cubic' :: Find the maxima and minima by fitting a cubic spline and
                    finding the roots of its derivative.
                    Expect input as xr.DataArrays
        DB NOTE: Currently only the 'comp' and 'cubic' method are implemented.
                Future methods include linear interpolation.

        JP NOTE: Cubic method:
            i) has intelligent fix for NaNs,

    Example usage:
    see example_scripts/tidegauge_tutorial.py
    """

    if method == "cubic":
        if (type(x) != xr.DataArray) or (type(y) != xr.DataArray):
            msg = "With method {} require input to be type: xr.DataArray" + " not {} and {}\n Reset method as comp"
            print(msg.format(method, type(x), type(y)))
            method = "comp"

    if method == "comp":
        peaks, props = scipy.signal.find_peaks(np.copy(y), **kwargs)
        return x[peaks], y[peaks]

    if method == "cubic":
        """
        Cleverness found at
        https://stackoverflow.com/questions/50371298/find-maximum-minimum-of-a-1d-interpolated-function

        Find the extrema on a cubic spline fitted to 1d array. E.g:
        # Some data
        x_axis = xr.DataArray([ 2.14414414,  2.15270826,  2.16127238,  2.1698365 ,  2.17840062, 2.18696474,  2.19552886,  2.20409298,  2.2126571 ,  2.22122122])
        y_axis = xr.DataArray([ 0.67958442,  0.89628424,  0.78904004,  3.93404167,  6.46422317, 6.40459954,  3.80216674,  0.69641825,  0.89675386,  0.64274198])
        # Fit cubic spline
        f = scipy.interpolate.interp1d(x_axis, y_axis, kind = 'cubic')
        x_new = np.linspace(x_axis[0], x_axis[-1],100)
        cr_pts, cr_vals = stats_utils.find_maxima(x_axis, y_axis, method='cubic')
        fig = plt.subplots()
        plt.plot(x_axis, y_axis, 'r+') # The fitted spline
        plt.plot(x_new, f(x_new)) # The fitted spline
        plt.plot(cr_pts, cr_vals, 'o') # The extrema

        """
        # Remove NaNs
        nan_mask = np.isnan(y)
        if sum(nan_mask) > 0:
            print("find_maxima(): There were NaNs in timeseries")
            x = x[np.logical_not(nan_mask)]
            y = y[np.logical_not(nan_mask)]

        # Sort over time. Monotonic increasing
        y = y.sortby(x)
        x = x.sortby(x)

        # Convert x to float64 (assuming y is/similar to np.float64)
        if type(x.values[0]) == np.datetime64:  # convert to decimal sec since 1970
            x_float = ((x.values - np.datetime64("1970-01-01T00:00:00")) / np.timedelta64(1, "s")).astype("float64")
            # x_float = x.values.astype('float64')
            y_float = y.values.astype("float64")
            flag_dt64 = True
        else:
            x_float = x.values.astype("float64")
            y_float = y.values.astype("float64")
            flag_dt64 = False

        if type(y.values[0]) != np.float64:
            print("find_maxima(): type(y)=", type(y))
            print("I was expecting a np.float64")

        # Do the interpolation
        f = scipy.interpolate.InterpolatedUnivariateSpline(x_float, y, k=3)
        # Find the extrema as roots
        extr_x_vals = quadratic_spline_roots(f.derivative())  # x values of extrema
        # Find the maxima roots
        extr_x_vals = np.hstack(
            [x_float[0], extr_x_vals, x_float[-1]]
        )  # add buffer points to ensure extrema are within
        ind = scipy.signal.argrelmax(f(extr_x_vals))[0]  # index that gives max(f) over extrema x locations
        max_vals = f(extr_x_vals[ind])

        # Convert back to datetime64 if appropriate
        y_out = max_vals
        if flag_dt64:
            N = len(extr_x_vals[ind])
            x_out = [
                np.datetime64("1970-01-01T00:00:00") + np.timedelta64(int(extr_x_vals[ind[i]]), "s") for i in range(N)
            ]
        else:
            x_out = extr_x_vals[ind]

        # restore xarray structure
        new_x = xr.DataArray(x_out, coords=[x_out], dims=x.dims)
        new_x.name = x.name
        new_y = xr.DataArray(y_out, coords=[x_out], dims=y.dims)
        new_y.name = y.name

        return new_x, new_y


def doodson_x0_filter(elevation, ax=0):
    """
    The Doodson X0 filter is a simple filter designed to damp out the main
    tidal frequencies. It takes hourly values, 19 values either side of the
    central one and applies a weighted average using:
              (1010010110201102112 0 2112011020110100101)/30.
    ( http://www.ntslf.org/files/acclaimdata/gloup/doodson_X0.html )

    In "Data Analaysis and Methods in Oceanography":

    "The cosine-Lanczos filter, the transform filter, and the
    Butterworth filter are often preferred to the Godin filter,
    to earlier Doodson filter, because of their superior ability
    to remove tidal period variability from oceanic signals."

    This routine can be used for any dimension input array.

    Parameters
    ----------
        elevation (ndarray) : Array of hourly elevation values.
        axis (int) : Time axis of input array. This axis must have >= 39
        elements

    Returns
    -------
        Filtered array of same rank as elevation.
    """
    if elevation.shape[ax] < 39:
        error("Doodson_XO: Ensure time axis has >=39 elements. Returning.")
        return
    # Define DOODSON XO weights
    kern = np.array(
        [
            1,
            0,
            1,
            0,
            0,
            1,
            0,
            1,
            1,
            0,
            2,
            0,
            1,
            1,
            0,
            2,
            1,
            1,
            2,
            0,
            2,
            1,
            1,
            2,
            0,
            1,
            1,
            0,
            2,
            0,
            1,
            1,
            0,
            1,
            0,
            0,
            1,
            0,
            1,
        ]
    )
    kern = kern / 30

    # Convolve input array with weights along the specified axis.
    filtered = np.apply_along_axis(lambda m: np.convolve(m, kern, mode=1), axis=ax, arr=elevation)

    # Pad out boundary areas with NaNs for given (arbitrary) axis.
    # DB: Is this the best way to do this?? Can put_along_axis be used instead
    filtered = filtered.swapaxes(0, ax)
    filtered[:19] = np.nan
    filtered[-19:] = np.nan
    filtered = filtered.swapaxes(0, ax)
    return filtered



"""Tide Gauge class"""
from .timeseries import Timeseries
import numpy as np
import xarray as xr
import matplotlib.pyplot as plt
import pandas as pd
import glob
import re
import pytz
from .._utils import general_utils, plot_util, crps_util, stats_util
from .._utils.logging_util import get_slug, debug, error, info
from typing import Union
from pathlib import Path


class Tidegauge(Timeseries):
    """
    This is an object for storage and manipulation of tide gauge data
    in a single dataset. This may require some processing of the observations
    such as interpolation to a common time step.

    This object's dataset should take the form (as with Timeseries):

        Dimensions:
            id_dim   : The locations dimension. Each time series has an index
            time : The time dimension. Each datapoint at each port has an index

        Coordinates:
            longitude (id_dim) : Longitude values for each port index
            latitude  (id_dim) : Latitude values for each port index
            time      (time) : Time values for each time index (datetime)
            id_name   (id_dim)   : Name of index, e.g. port name or mooring id.

    An example data variable could be ssh, or ntr (non-tidal residual). This
    object can also be used for other instrument types, not just tide gauges.
    For example moorings.

    Every id index for this object should use the same time coordinates.
    Therefore, timeseries need to be aligned before being placed into the
    object. If there is any padding needed, then NaNs should be used. NaNs
    should also be used for quality control/data rejection.
    """

    def __init__(self, dataset=None, config: Union[Path, str] = None, new_time_coords=None):
        """
        Initialise TIDEGAUGE object as empty or by providing an existing
        dataset or tidegauge object. There are read functions within Tidegauge()
        which can subsequently read into this object's dataset.'

        Example usage:
        --------------
        # Create empty tidegauge object
        tidegauge = coast.Tidegauge()

        # Create new tidegauge object containing an existing xarray dataset
        tidegauge = coast.Tidegauge(dataset = dataset_name)

        # Create new tidegauge object containing existing xarray dataset and
        # create new time coordinate (will remove all variables)
        tidegauge = coast.Tidegage(tidegauge0.dataset, new_coords = time_array)

        INPUTS
        ----------
        dataset (xr.Dataset)    :: Xarray dataset to insert into new object
        config  (Path or Str)   :: Configuration file to use when calling
                                   read functions (saved in object) [Optional]
        new_time_coords (array) :: New time coords to create [Optional]

        Returns
        -------
        New Tidegauge object.
        """
        debug(f"Creating a new ..... {get_slug(self)}")
        super().__init__(config)

        # If file list is supplied, read files from directory
        if dataset is not None:
            self.dataset = dataset.copy()
            self.apply_config_mappings()

            # If new_time_coords, replace existing time dimension with new
            if new_time_coords is not None:
                ds_coords = xr.Dataset(self.dataset.coords).drop("time")
                ds_coords["time"] = ("t_dim", new_time_coords)
                ds_coords = ds_coords.set_coords("time")
                self.dataset = ds_coords.copy()

        else:
            self.dataset = None

        print(f"{get_slug(self)} initialised")

    ############ tide gauge methods ###########################################
    def read_gesla_v3(self, fn_gesla, date_start=None, date_end=None):
        """
        For reading from a GESLA2 (Format version 3.0) file(s) into an
        xarray dataset. Formatting according to Woodworth et al. (2017).
        Website: https://www.gesla.org/
        If no data lies between the specified dates, a dataset is still created
        containing information on the tide gauge, but the time dimension will
        be empty.
        Parameters
        ----------
        fn_gesla (str) : path to gesla tide gauge file, list of files or a glob
        date_start (datetime) : start date for returning data
        date_end (datetime) : end date for returning data

        Returns
        -------
        Creates xarray.dataset within tidegauge object containing loaded data.
        If multiple files are provided then instead returns a list of NEW
        tidegauge objects.
        """
        debug(f'Reading "{fn_gesla}" as a GESLA file with {get_slug(self)}')
        # TODO Maybe include start/end dates
        dataset = None

        # See if its a file list input, or a glob
        if type(fn_gesla) is not list:
            file_list = glob.glob(fn_gesla)
        else:
            file_list = fn_gesla

        multiple = False
        if len(file_list) > 1:
            multiple = True

        ds_list = []
        # Loop over files and put resulting dataset into an output list
        for fn in file_list:
            try:
                header_dict = self._read_gesla_header_v3(fn)
                dataset = self._read_gesla_data_v3(fn, date_start, date_end)
            except:
                raise Exception("Problem reading GESLA file: " + fn)
            # Attributes
            dataset["longitude"] = ("id_dim", [header_dict["longitude"]])
            dataset["latitude"] = ("id_dim", [header_dict["latitude"]])
            dataset["id_name"] = ("id_dim", [header_dict["site_name"]])
            dataset = dataset.set_coords(["longitude", "latitude", "id_name"])

            # Create tidegauge object, save dataset and append to list
            if multiple:
                tg_tmp = Tidegauge()
                tg_tmp.dataset = dataset
                tg_tmp.apply_config_mappings()
                ds_list.append(tg_tmp)

        # If there is only one file, then just return the dataset, not a list
        if multiple:
            return ds_list
        else:
            self.dataset = dataset
            self.apply_config_mappings()

    @classmethod
    def _read_gesla_header_v3(cls, fn_gesla):
        """
        Reads header from a GESLA file (format version 3.0).

        Parameters
        ----------
        fn_gesla (str) : path to gesla tide gauge file

        Returns
        -------
        dictionary of attributes
        """
        debug(f'Reading GESLA header from "{fn_gesla}"')
        fid = open(fn_gesla)

        # Read lines one by one (hopefully formatting is consistent)
        fid.readline()  # Skip first line
        # Geographical stuff
        site_name = fid.readline().split()[3:]
        site_name = "_".join(site_name)
        country = fid.readline().split()[2:]
        country = "_".join(country)
        contributor = fid.readline().split()[2:]
        contributor = "_".join(contributor)
        # Coordinates
        latitude = float(fid.readline().split()[2])
        longitude = float(fid.readline().split()[2])
        coordinate_system = fid.readline().split()[3]
        # Dates
        start_date = fid.readline().split()[3:5]
        start_date = " ".join(start_date)
        start_date = pd.to_datetime(start_date)
        end_date = fid.readline().split()[3:5]
        end_date = " ".join(end_date)
        end_date = pd.to_datetime(end_date)
        time_zone_hours = float(fid.readline().split()[4])
        # Other
        fid.readline()  # Datum
        fid.readline()  # Instrument
        precision = float(fid.readline().split()[2])
        null_value = float(fid.readline().split()[3])

        debug(f'Read done, close file "{fn_gesla}"')
        fid.close()
        # Put all header info into an attributes dictionary
        header_dict = {
            "site_name": site_name,
            "country": country,
            "contributor": contributor,
            "latitude": latitude,
            "longitude": longitude,
            "coordinate_system": coordinate_system,
            "original_start_date": start_date,
            "original_end_date": end_date,
            "time_zone_hours": time_zone_hours,
            "precision": precision,
            "null_value": null_value,
        }
        return header_dict

    @classmethod
    def _read_gesla_data_v3(cls, fn_gesla, date_start=None, date_end=None, header_length: int = 32):
        """
        Reads observation data from a GESLA file (format version 3.0).

        Parameters
        ----------
        fn_gesla (str) : path to gesla tide gauge file
        date_start (datetime) : start date for returning data
        date_end (datetime) : end date for returning data
        header_length (int) : number of lines in header (to skip when reading)

        Returns
        -------
        xarray.Dataset containing times, sealevel and quality control flags
        """
        # Initialise empty dataset and lists
        debug(f'Reading GESLA data from "{fn_gesla}"')
        dataset = xr.Dataset()
        time = []
        ssh = []
        qc_flags = []
        # Open file and loop until EOF
        with open(fn_gesla) as file:
            line_count = 1
            for line in file:
                # Read all data. Date boundaries are set later.
                if line_count > header_length:
                    working_line = line.split()
                    if working_line[0] != "#":
                        time.append(working_line[0] + " " + working_line[1])
                        ssh.append(float(working_line[2]))
                        qc_flags.append(int(working_line[3]))

                line_count = line_count + 1
            debug(f'Read done, close file "{fn_gesla}"')

        # Convert time list to datetimes using pandas
        time = np.array(pd.to_datetime(time))

        # Return only values between stated dates
        start_index = 0
        end_index = len(time)
        if date_start is not None:
            date_start = np.datetime64(date_start)
            start_index = np.argmax(time >= date_start)
        if date_end is not None:
            date_end = np.datetime64(date_end)
            end_index = np.argmax(time > date_end)
        time = time[start_index:end_index]
        ssh = ssh[start_index:end_index]
        qc_flags = qc_flags[start_index:end_index]

        # Set null values to nan
        ssh = np.array(ssh)
        qc_flags = np.array(qc_flags)
        ssh[qc_flags == 5] = np.nan

        # Assign arrays to Dataset
        dataset["ssh"] = xr.DataArray(ssh, dims=["t_dim"]).expand_dims("id_dim")
        dataset["qc_flags"] = xr.DataArray(qc_flags, dims=["t_dim"]).expand_dims("id_dim")
        dataset = dataset.assign_coords(time=("t_dim", time))

        # Assign local dataset to object-scope dataset
        return dataset

    ### tide table methods (HLW)
    def read_hlw(self, fn_hlw, date_start=None, date_end=None):
        """
        For reading from a file of tidetable High and Low Waters (HLW) data into an
        xarray dataset. File contains high water and low water heights and times

        If no data lies between the specified dates, a dataset is still created
        containing information on the tide gauge, but the time dimension will
        be empty.

        The data takes the form:
        LIVERPOOL (GLADSTONE DOCK)    TZ: UT(GMT)/BST     Units: METRES    Datum: Chart Datum
        01/10/2020  06:29    1.65
        01/10/2020  11:54    9.01
        01/10/2020  18:36    1.87
        ...

        Parameters
        ----------
        fn_hlw (str) : path to tabulated High Low Water file
        date_start (datetime) : start date for returning data
        date_end (datetime) : end date for returning data

        Returns
        -------
        xarray.Dataset object.
        """
        debug(f'Reading "{fn_hlw}" as a HLW file with {get_slug(self)}')
        # TODO Maybe include start/end dates
        try:
            header_dict = self._read_hlw_header(fn_hlw)
            dataset = self._read_hlw_data(fn_hlw, header_dict, date_start, date_end)
            if header_dict["field"] == "TZ:UT(GMT)/BST":
                debug("Read in as BST, stored as UTC")
            elif header_dict["field"] == "TZ:GMTonly":
                debug("Read and store as GMT/UTC")
            else:
                debug("Not expecting that timezone")

        except:
            raise Exception("Problem reading HLW file: " + fn_hlw)

        dataset.attrs = header_dict
        self.dataset = dataset
        self.apply_config_mappings()

    @classmethod
    def _read_hlw_header(cls, filnam):
        """
        Reads header from a HWL file.

        The data takes the form:
        LIVERPOOL (GLADSTONE DOCK) TZ: UT(GMT)/BST Units: METRES Datum: Chart Datum
        01/10/2020  06:29    1.65
        01/10/2020  11:54    9.01
        01/10/2020  18:36    1.87
        ...

        Parameters
        ----------
        filnam (str) : path to file

        Returns
        -------
        dictionary of attributes
        """
        debug(f'Reading HLW header from "{filnam}" ')
        fid = open(filnam)

        # Read lines one by one (hopefully formatting is consistent)
        header = re.split(r"\s{2,}", fid.readline())
        site_name = header[0]
        site_name = site_name.replace(" ", "")

        field = header[1]
        field = field.replace(" ", "")

        units = header[2]
        units = units.replace(" ", "")

        datum = header[3]
        datum = datum.replace(" ", "")

        debug(f'Read done, close file "{filnam}"')
        fid.close()
        # Put all header info into an attributes dictionary
        header_dict = {"site_name": site_name, "field": field, "units": units, "datum": datum}
        return header_dict

    @classmethod
    def _read_hlw_data(cls, filnam, header_dict, date_start=None, date_end=None, header_length: int = 1):
        """
        Reads HLW data from a tidetable file.

        Parameters
        ----------
        filnam (str) : path to HLW tide gauge file
        date_start (np.datetime64) : start date for returning data.
        date_end (np.datetime64) : end date for returning data.
        header_length (int) : number of lines in header (to skip when reading)

        Returns
        -------
        xarray.Dataset containing times, High and Low water values
        """
        import datetime

        # Initialise empty dataset and lists
        debug(f'Reading HLW data from "{filnam}"')
        dataset = xr.Dataset()
        time = []
        ssh = []

        if header_dict["field"] == "TZ:UT(GMT)/BST":
            localtime_flag = True
        else:
            localtime_flag = False

        # Open file and loop until EOF
        with open(filnam) as file:
            line_count = 1
            for line in file:
                # Read all data. Date boundaries are set later.
                if line_count > header_length:
                    working_line = line.split()
                    if working_line[0] != "#":
                        time_str = working_line[0] + " " + working_line[1]
                        # Read time as datetime.datetime because it can handle local timezone easily AND the unusual date format
                        datetime_obj = datetime.datetime.strptime(time_str, "%d/%m/%Y %H:%M")
                        if localtime_flag == True:
                            bst_obj = pytz.timezone("Europe/London")
                            time.append(np.datetime64(bst_obj.localize(datetime_obj).astimezone(pytz.utc)))
                        else:
                            time.append(np.datetime64(datetime_obj))
                        ssh.append(float(working_line[2]))
                line_count = line_count + 1
            debug(f'Read done, close file "{filnam}"')

        # Return only values between stated dates
        start_index = 0
        end_index = len(time)
        if date_start is not None:
            # start_index = general_utils.nearest_datetime_ind(time, date_start)
            date_start = np.datetime64(date_start)
            start_index = np.argmax(time >= date_start)
            debug(f"date_start: {date_start}. start_index: {start_index}")
        if date_end is not None:
            # end_index = general_utils.nearest_datetime_ind(time, date_end)
            date_end = np.datetime64(date_end)
            end_index = np.argmax(time > date_end)
            debug(f"date_end: {date_end}. end_index: {end_index}")
        time = time[start_index:end_index]
        ssh = ssh[start_index:end_index]
        debug(f"ssh: {ssh}")
        # Assign arrays to Dataset
        dataset["ssh"] = xr.DataArray(ssh, dims=["time"]).expand_dims("id_dim")
        dataset = dataset.assign_coords(time=("time", time))
        # Assign local dataset to object-scope dataset
        return dataset

    def show(self, timezone: str = None):
        """
        Print out the values in the xarray
        Displays with specified timezone
        """
        # debug(" Saltney pred", np.datetime_as_string(Saltney_time_pred[i], unit='m', timezone=pytz.timezone('Europe/London')),". Height: {:.2f} m".format( HT.values[i] ))
        if timezone == None:
            for i in range(len(self.dataset.ssh)):
                #               debug('time:', self.dataset.time[i].values,
                debug(
                    "time (UTC):",
                    general_utils.dayoweek(self.dataset.time[i].values),
                    np.datetime_as_string(self.dataset.time[i], unit="m"),
                    "height:",
                    self.dataset.ssh[i].values,
                    "m",
                )
        else:  # display timezone aware times
            for i in range(len(self.dataset.ssh)):
                #               debug('time:', self.dataset.time[i].values,
                debug(
                    "time (" + timezone + "):",
                    general_utils.day_of_week(self.dataset.time[i].values),
                    np.datetime_as_string(self.dataset.time[i], unit="m", timezone=pytz.timezone(timezone)),
                    "height:",
                    self.dataset.ssh[i].values,
                    "m",
                )

    def get_tide_table_times(
        self,
        time_guess: np.datetime64 = None,
        time_var: str = "time",
        measure_var: str = "ssh",
        method: str = "window",
        winsize=None,
    ):
        """
        Get tide times and heights from tide table.
        input:
        time_guess : np.datetime64 or datetime
                assumes utc
        time_var : name of time variable [default: 'time']
        measure_var : name of ssh variable [default: 'ssh']

        method =
            window:  +/- hours window size, winsize, (int) return values in that window
                uses additional variable winsize (int) [default 2hrs]
            nearest_1: return only the nearest event, if in winsize [default:None]
            nearest_2: return nearest event in future and the nearest in the past (i.e. high and a low), if in winsize [default:None]
            nearest_HW: return nearest High Water event (computed as the max of `nearest_2`), if in winsize [default:None]

        returns: xr.DataArray( measure_var, coords=time_var)
            E.g. ssh (m), time (utc)
            If value is not found, it returns a NaN with time value as the
            guess value.

        """
        # Ensure the date objects are datetime
        if type(time_guess) is not np.datetime64:
            debug("Convert date to np.datetime64")
            time_guess = np.datetime64(time_guess)

        if time_guess == None:
            debug("Use today's date")
            time_guess = np.datetime64("now")

        if method == "window":
            if winsize == None:
                winsize = 2
            # initialise start_index and end_index
            start_index = 0
            end_index = len(self.dataset[time_var])

            date_start = time_guess - np.timedelta64(winsize, "h")
            start_index = np.argmax(self.dataset[time_var].values >= date_start)

            date_end = time_guess + np.timedelta64(winsize, "h")
            end_index = np.argmax(self.dataset[time_var].values > date_end)

            ssh = self.dataset[measure_var].isel(time=slice(start_index, end_index))

            return ssh[0]

        elif method == "nearest_1":
            dt = np.abs(self.dataset[time_var] - time_guess)
            index = np.argsort(dt).values
            if winsize is not None:  # if search window trucation exists
                if np.timedelta64(dt[index[0]].values, "m").astype("int") <= 60 * winsize:  # compare in minutes
                    debug(f"dt:{np.timedelta64(dt[index[0]].values, 'm').astype('int')}")
                    debug(f"winsize:{winsize}")
                    return self.dataset[measure_var].isel(time=index[0])
                else:
                    # return a NaN in an xr.Dataset
                    # The rather odd trailing zero is to remove the array layer
                    # on both time and measurement, and to match the other
                    # alternative for a return object
                    return xr.DataArray([np.NaN], dims=(time_var), coords={time_var: [time_guess]})[0]
            else:  # give the closest without window search truncation
                return self.dataset[measure_var].isel(time=index[0])

        elif method == "nearest_2":
            index = np.argsort(np.abs(self.dataset[time_var] - time_guess)).values
            nearest_2 = self.dataset[measure_var].isel(time=index[0 : 1 + 1])  # , self.dataset.time[index[0:1+1]]
            return nearest_2[0]

        elif method == "nearest_HW":
            index = np.argsort(np.abs(self.dataset[time_var] - time_guess)).values
            # return self.dataset.ssh[ index[np.argmax( self.dataset.ssh[index[0:1+1]]] )] #, self.dataset.time[index[0:1+1]]
            nearest_2 = self.dataset[measure_var].isel(time=index[0 : 1 + 1])  # , self.dataset.time[index[0:1+1]]
            return nearest_2.isel(time=nearest_2.argmax())

        else:
            debug("Not expecting that option / method")

    ############ environment.data.gov.uk gauge methods ###########################
    @classmethod
    def read_ea_api_to_xarray(
        cls, n_days: int = 5, date_start: np.datetime64 = None, date_end: np.datetime64 = None, station_id="E70124"
    ):
        """
        load gauge data via environment.data.gov.uk EA API
        Either loads last n_days, or from date_start:date_end

        API Source:
        https://environment.data.gov.uk/flood-monitoring/doc/reference

        Details of available tidal stations are recovered with:
        https://environment.data.gov.uk/flood-monitoring/id/stations?type=TideGauge
        Recover the "stationReference" for the gauge of interest and pass as
        station_id:str. The default station_id="E70124" is Liverpool.

        INPUTS:
            n_days : int. Extact the last n_days from now.
            date_start : datetime. UTC format string "yyyy-MM-dd" E.g 2020-01-05
            date_end : datetime
            station_id : int. Station id. Also referred to as stationReference in
             EA API. Default value is for Liverpool.
        OUTPUT:
            ssh, time : xr.Dataset
        """
        import requests, json

        cls.n_days = n_days
        cls.date_start = date_start
        cls.date_end = date_end
        cls.station_id = station_id  # EA id: stationReference

        # Obtain and process header information
        info("load station info")
        url = "https://environment.data.gov.uk/flood-monitoring/id/stations/" + cls.station_id + ".json"
        try:
            request_raw = requests.get(url)
            header_dict = json.loads(request_raw.content)
        except ValueError:
            debug(f"Failed request for station {cls.station_id}")
            return

        try:
            header_dict["site_name"] = header_dict["items"]["label"]
            header_dict["latitude"] = header_dict["items"]["lat"]
            header_dict["longitude"] = header_dict["items"]["long"]
        except:
            info(f"possible missing some header info: site_name,latitude,longitude")
        try:
            # Define url call with parameter from station info
            htmlcall_station_id = header_dict["items"]["measures"]["@id"] + "/readings?"
        except:
            debug(f"problem defining the parameter to read")

        # Construct API request for data recovery
        info("load station data")
        if (cls.date_start == None) & (cls.date_end == None):
            info(f"GETting n_days= {cls.n_days} of data")
            url = (
                htmlcall_station_id
                + "since="
                + (np.datetime64("now") - np.timedelta64(n_days, "D")).item().strftime("%Y-%m-%dT%H:%M:%SZ")
            )
            debug(f"url request: {url}")
        else:
            # Check date_start and date_end are timetime objects
            if (type(cls.date_start) is np.datetime64) & (type(cls.date_end) is np.datetime64):
                info(f"GETting data from {cls.date_start} to {cls.date_end}")
                startdate = cls.date_start.item().strftime("%Y-%m-%d")
                enddate = cls.date_end.item().strftime("%Y-%m-%d")
                url = htmlcall_station_id + "startdate=" + startdate + "&enddate=" + enddate
                debug(f"url request: {url}")

            else:
                debug("Expecting date_start and date_end as datetime objects")

        # Get the data
        try:
            request_raw = requests.get(url)
            request = json.loads(request_raw.content)
            debug(f"EA API request: {request_raw.text}")
        except ValueError:
            debug(f"Failed request: {request_raw}")
            return

        # Process timeseries data
        dataset = xr.Dataset()
        time = []
        ssh = []
        nvals = len(request["items"])
        time = np.array([np.datetime64(request["items"][i]["dateTime"]) for i in range(nvals)])
        ssh = np.array([request["items"][i]["value"] for i in range(nvals)])

        # Assign arrays to Dataset
        dataset["ssh"] = xr.DataArray(ssh, dims=["time"])
        dataset = dataset.assign_coords(time=("time", time))
        dataset.attrs = header_dict
        debug(f"EA API request headers: {header_dict}")
        # debug(f"EA API request 1st time: {time[0]} and value: {ssh[0]}")

        # Assign local dataset to object-scope dataset
        return dataset

    ############ BODC tide gauge methods ######################################
    def read_bodc(self, fn_bodc, date_start=None, date_end=None):
        """
        For reading from a single BODC (processed) file into an
        xarray dataset.
        If no data lies between the specified dates, a dataset is still created
        containing information on the tide gauge, but the time dimension will
        be empty.

        Data name: UK Tide Gauge Network, processed data.
        Source: https://www.bodc.ac.uk/
        See data notes from source for description of QC flags.

        The data takes the form:
            Port:              P234
            Site:              Liverpool, Gladstone Dock
            Latitude:          53.44969
            Longitude:         -3.01800
            Start Date:        01AUG2020-00.00.00
            End Date:          31AUG2020-23.45.00
            Contributor:       National Oceanography Centre, Liverpool
            Datum information: The data refer to Admiralty Chart Datum (ACD)
            Parameter code:    ASLVBG02 = Surface elevation (unspecified datum)
            of the water body by bubbler tide gauge (second sensor)
              Cycle    Date      Time    ASLVBG02   Residual
             Number yyyy mm dd hh mi ssf         f          f
                 1) 2020/08/01 00:00:00     5.354M     0.265M
                 2) 2020/08/01 00:15:00     5.016M     0.243M
                 3) 2020/08/01 00:30:00     4.704M     0.241M
                 4) 2020/08/01 00:45:00     4.418M     0.255M
                 5) 2020/08/01 01:00:00     4.133      0.257
                 ...

        Parameters
        ----------
        fn_bodc (str) : path to bodc tide gauge file
        date_start (datetime) : start date for returning data
        date_end (datetime) : end date for returning data

        Returns
        -------
        xarray.Dataset object.
        """
        debug(f'Reading "{fn_bodc}" as a BODC file with {get_slug(self)}')
        # TODO Maybe include start/end dates
        try:
            header_dict = self._read_bodc_header(fn_bodc)
            dataset = self._read_bodc_data(fn_bodc, date_start, date_end)
        except:
            raise Exception("Problem reading BODC file: " + fn_bodc)
        # Attributes
        dataset["longitude"] = header_dict["longitude"]
        dataset["latitude"] = header_dict["latitude"]
        del header_dict["longitude"]
        del header_dict["latitude"]

        dataset.attrs = header_dict
        self.dataset = dataset

    @staticmethod
    def _read_bodc_header(fn_bodc):
        """
        Reads header from a BODC file (format version 3.0).

        Parameters
        ----------
        fn_bodc (str) : path to bodc tide gauge file

        Returns
        -------
        dictionary of attributes
        """
        debug(f'Reading BODC header from "{fn_bodc}"')
        fid = open(fn_bodc)

        # Read lines one by one (hopefully formatting is consistent)
        # Geographical stuff
        header_dict = {}
        header = True
        for line in fid:
            if ":" in line and header == True:
                (key, val) = line.split(":")
                key = key.lower().strip().replace(" ", "_")
                val = val.lower().strip().replace(" ", "_")
                header_dict[key] = val
                debug(f"Header key: {key} and value: {val}")
            else:
                # debug('No colon')
                header = False
        header_dict["site_name"] = header_dict["site"]  # duplicate as standard name
        debug(f'Read done, close file "{fn_bodc}"')
        fid.close()

        header_dict["latitude"] = float(header_dict["latitude"])
        header_dict["longitude"] = float(header_dict["longitude"])

        return header_dict

    @staticmethod
    def _read_bodc_data(fn_bodc, date_start=None, date_end=None, header_length: int = 11):
        """
        Reads observation data from a BODC file.

        Parameters
        ----------
        fn_bodc (str) : path to bodc tide gauge file
        date_start (datetime) : start date for returning data
        date_end (datetime) : end date for returning data
        header_length (int) : number of lines in header (to skip when reading)

        Returns
        -------
        xarray.Dataset containing times, sealevel and quality control flags
        """
        # Initialise empty dataset and lists
        debug(f'Reading BODC data from "{fn_bodc}"')
        dataset = xr.Dataset()
        time = []
        ssh = []
        qc_flags = []
        residual = []
        # Open file and loop until EOF
        with open(fn_bodc) as file:
            line_count = 1
            for line in file:
                # Read all data. Date boundaries are set later.
                if line_count > header_length:
                    try:
                        working_line = line.split()
                        time_str = working_line[1] + " " + working_line[2]  # Empty lines cause trouble
                        ssh_str = working_line[3]
                        residual_str = working_line[4]
                        if ssh_str[-1].isalpha():
                            qc_flag_str = ssh_str[-1]
                            ssh_str = ssh_str.replace(qc_flag_str, "")
                            residual_str = residual_str.replace(qc_flag_str, "")
                        elif residual_str[-1].isalpha():  # sometimes residual has a
                            # flag when elevation does not
                            qc_flag_str = residual_str[-1]
                            ssh_str = ssh_str.replace(qc_flag_str, "")
                            residual_str = residual_str.replace(qc_flag_str, "")
                        else:
                            qc_flag_str = ""
                        # debug(line_count-header_length, residual_str, float(residual_str))
                        # debug(working_line, ssh_str, qc_flag_str)
                        time.append(time_str)
                        qc_flags.append(qc_flag_str)
                        ssh.append(float(ssh_str))
                        residual.append(float(residual_str))
                    except:
                        debug(f"{file} probably empty line at end. Breaks split()")
                line_count = line_count + 1
            debug(f'Read done, close file "{fn_bodc}"')

        # Convert time list to datetimes using pandas
        time = np.array(pd.to_datetime(time))

        # Return only values between stated dates
        start_index = 0
        end_index = len(time)
        if date_start is not None:
            date_start = np.datetime64(date_start)
            start_index = np.argmax(time >= date_start)
        if date_end is not None:
            date_end = np.datetime64(date_end)
            end_index = np.argmax(time > date_end)
        time = time[start_index:end_index]
        ssh = ssh[start_index:end_index]
        qc_flags = qc_flags[start_index:end_index]

        # Set null values to nan
        ssh = np.array(ssh)
        qc_flags = np.array(qc_flags)
        # ssh[qc_flags==5] = np.nan

        # Assign arrays to Dataset
        dataset["ssh"] = xr.DataArray(ssh, dims=["time"]).expand_dims("id_dim")
        dataset["qc_flags"] = xr.DataArray(qc_flags, dims=["time"]).expand_dims("id_dim")
        dataset = dataset.assign_coords(time=("time", time))

        # Assign local dataset to object-scope dataset
        return dataset

    ##############################################################################
    ###                ~            Plotting             ~                     ###
    ##############################################################################

    def plot_timeseries(self, id, var_list=["ssh"], date_start=None, date_end=None, plot_line=False):
        """
        Quick plot of time series stored within object's dataset
        Parameters
        ----------
        date_start (datetime) : Start date for plotting
        date_end (datetime) : End date for plotting
        var_list  (str)  : List of variables to plot. Default: just ssh
        plot_line (bool) : If true, draw line between markers

        Returns
        -------
        matplotlib figure and axes objects
        """
        debug(f"Plotting timeseries for {get_slug(self)}")
        fig = plt.figure(figsize=(10, 10))
        # Check input is a list (even for one variable)
        if type(var_list) is str:
            var_list = [var_list]

        for var_str in var_list:
            dim_str = self.dataset[var_str].dims[0]
            x = np.array(self.dataset[dim_str])
            y = np.array(self.dataset[var_str])

            # Use only values between stated dates
            start_index = 0
            end_index = len(x)
            if date_start is not None:
                date_start = np.datetime64(date_start)
                start_index = np.argmax(x >= date_start)
            if date_end is not None:
                date_end = np.datetime64(date_end)
                end_index = np.argmax(x > date_end)
            x = x[start_index:end_index]
            y = y[start_index:end_index]

            # Plot lines first if needed
            if plot_line:
                plt.plot(x, y, c=[0.5, 0.5, 0.5], linestyle="--", linewidth=0.5)

            ax = plt.scatter(x, y, s=10)

        plt.grid()
        plt.xticks(rotation=45)
        plt.legend(var_list)
        # Title and axes
        plt.xlabel("Date")
        plt.title("Site: " + self.dataset.site_name)

        return fig, ax

    def plot_on_map(self):
        """
        Show the location of a tidegauge on a map.
        Example usage:
        --------------
        # For a TIDEGAUGE object tg
        tg.plot_map()
        """

        debug(f"Plotting tide gauge locations for {get_slug(self)}")

        X = self.dataset.longitude
        Y = self.dataset.latitude
        fig, ax = plot_util.geo_scatter(X, Y)
        ax.set_xlim((X - 10, X + 10))
        ax.set_ylim((Y - 10, Y + 10))
        return fig, ax

    @classmethod
    def plot_on_map_multiple(cls, tidegauge_list, color_var_str=None):
        """
        Show the location of a tidegauge on a map.
        Example usage:
        --------------
        # For a TIDEGAUGE object tg
        tg.plot_map()
        """

        debug(f"Plotting tide gauge locations for {get_slug(cls)}")

        X = []
        Y = []
        C = []
        for tg in tidegauge_list:
            X.append(tg.dataset.longitude)
            Y.append(tg.dataset.latitude)
            if color_var_str is not None:
                C.append(tg.dataset[color_var_str].values)

        title = ""
        if color_var_str is None:
            fig, ax = plot_util.geo_scatter(X, Y, title=title)
        else:
            fig, ax = plot_util.geo_scatter(X, Y, title=title, c=C)

        ax.set_xlim((min(X) - 10, max(X) + 10))
        ax.set_ylim((min(Y) - 10, max(Y) + 10))
        return fig, ax

    ##############################################################################
    ###                ~        Model Comparison         ~                     ###
    ##############################################################################
    def obs_operator(self, gridded, time_interp="nearest"):
        """
        Regrids a Gridded object onto a tidegauge_multiple object. A nearest
        neighbour interpolation is done for spatial interpolation and time
        interpolation can be specified using the time_interp argument. This
        takes any scipy interpolation string. If Gridded object contains a
        landmask variables, then the nearest WET point is taken for each tide
        gauge.

        Output is a new tidegauge_multiple object containing interpolated data.
        """

        gridded = gridded.dataset
        ds = self.dataset

        # Determine if landmask is present
        if "landmask" not in gridded:
            gridded.landmask = None

        # Determine spatial indices
        print("Calculating spatial indices.", flush=True)
        ind_x, ind_y = general_utils.nearest_indices_2d(
            gridded.longitude, gridded.latitude, ds.longitude, ds.latitude, mask=gridded.landmask
        )

        # Extract spatial time series
        print("Calculating time indices.", flush=True)
        extracted = gridded.isel(x_dim=ind_x, y_dim=ind_y)
        if "dim_0" in extracted.dims:
            extracted = extracted.swap_dims({"dim_0": "id_dim"})
        else:
            extracted = extracted.expand_dims("id_dim")

        # Compute data (takes a while..)
        print(" Indexing model data at tide gauge locations.. ", flush=True)
        extracted.load()

        # Check interpolation distances
        print("Calculating interpolation distances.", flush=True)
        interp_dist = general_utils.calculate_haversine_distance(
            extracted.longitude.values, extracted.latitude.values, ds.longitude.values, ds.latitude.values
        )

        # Interpolate model onto obs times
        print("Interpolating in time...", flush=True)
        extracted = extracted.rename({"time": "t_dim"})
        extracted = extracted.interp(t_dim=ds.time.values, method=time_interp)

        # Put interp_dist into dataset
        extracted["interp_dist"] = interp_dist
        extracted = extracted.rename_vars({"t_dim": "time"})

        tg_out = Tidegauge()
        tg_out.dataset = extracted
        return tg_out

    def time_slice(self, date0, date1):
        """Return new Gridded object, indexed between dates date0 and date1"""
        dataset = self.dataset
        t_ind = pd.to_datetime(dataset.time.values) >= date0
        dataset = dataset.isel(t_dim=t_ind)
        t_ind = pd.to_datetime(dataset.time.values) < date1
        dataset = dataset.isel(t_dim=t_ind)
        return Tidegauge(dataset=dataset)

    def subset_indices_lonlat_box(self, lonbounds, latbounds):
        """Get a subset of this Profile() object in a spatial box.

        lonbounds -- Array of form [min_longitude=-180, max_longitude=180]
        latbounds -- Array of form [min_latitude, max_latitude]

        return: A new profile object containing subsetted data
        """
        ind = general_utils.subset_indices_lonlat_box(
            self.dataset.longitude, self.dataset.latitude, lonbounds[0], lonbounds[1], latbounds[0], latbounds[1]
        )
        return Tidegauge(dataset=self.dataset.isel(id_dim=ind[0]))



"""A general utility file."""
import xarray as xr
import numpy as np
import sklearn.neighbors as nb
import pandas as pd


def determine_season(t):
    """
    Determine season (or array of seasons) from a time (Datetime or xarray)
    object. Put in an array of times, get out an array of seasons.
    """
    season_dict = {
        12: "DJF",
        1: "DJF",
        2: "DJF",
        3: "MAM",
        4: "MAM",
        5: "MAM",
        6: "JJA",
        7: "JJA",
        8: "JJA",
        9: "SON",
        10: "SON",
        11: "SON",
    }
    pd_month = pd.to_datetime(t).month
    pd_season = [season_dict[pp] for pp in pd_month]
    return np.array(pd_season)


def subset_indices_by_distance_balltree(longitude, latitude, centre_lon, centre_lat, radius: float, mask=None):
    """
    Returns the indices of points that lie within a specified radius (km) of
    central latitude and longitudes. This makes use of BallTree.query_radius.

    Parameters
    ----------
    longitude   : (numpy.ndarray) longitudes in degrees
    latitude    : (numpy.ndarray) latitudes in degrees
    centre_lon  : Central longitude. Can be single value or array of values
    centre_lat  : Central latitude. Can be single value or array of values
    radius      : (float) Radius in km within which to find indices
    mask        : (numpy.ndarray) of same dimension as longitude and latitude.
                  If specified, will mask out points from the routine.
    Returns
    -------
        Returns an array of indices corresponding to points within radius.
        If more than one central location is specified, this will be a list
        of index arrays. Each element of which corresponds to one centre.
    If longitude is 1D:
        Returns one array of indices per central location
    If longitude is 2D:
        Returns arrays of x and y indices per central location.
        ind_y corresponds to row indices of the original input arrays.
    """
    # change inputs to numpy
    longitude = np.array(longitude)
    latitude = np.array(latitude)
    centre_lon = np.array(centre_lon)
    centre_lat = np.array(centre_lat)
    # Calculate radius in radians
    earth_radius = 6371
    r_rad = radius / earth_radius
    # For reshaping indices at the end
    original_shape = longitude.shape
    # Check if radius centres are numpy arrays. If not, make them into ndarrays
    if not isinstance(centre_lon, np.ndarray):
        centre_lat = np.array(centre_lat)
        centre_lon = np.array(centre_lon)
    # Determine number of centres provided
    n_pts = 1 if centre_lat.shape == () else len(centre_lat)
    # If a mask is supplied, remove indices from arrays. Flatten input ready
    # for BallTree
    if mask is None:
        longitude = longitude.flatten()
        latitude = latitude.flatten()
    else:
        longitude[mask] = np.nan
        latitude[mask] = np.nan
        longitude = longitude.flatten()
        latitude = latitude.flatten()
    # Put lons and lats into 2D location arrays for BallTree: [lat, lon]
    locs = np.vstack((latitude, longitude)).transpose()
    locs = np.radians(locs)
    # Construct central input to BallTree.query_radius
    if n_pts == 1:
        centre = np.array([[centre_lat, centre_lon]])
    else:
        centre = np.vstack((centre_lat, centre_lon)).transpose()
    centre = np.radians(centre)
    # Do nearest neighbour interpolation using BallTree (gets indices)
    tree = nb.BallTree(locs, leaf_size=2, metric="haversine")
    ind_1d = tree.query_radius(centre, r=r_rad)
    if len(original_shape) == 1:
        return ind_1d
    else:
        # Get 2D indices from 1D index output from BallTree
        ind_y = []
        ind_x = []
        for ii in np.arange(0, n_pts):
            x_tmp, y_tmp = np.unravel_index(ind_1d[ii], original_shape)
            ind_x.append(x_tmp.squeeze())
            ind_y.append(y_tmp.squeeze())
        if n_pts == 1:
            return ind_x[0], ind_y[0]
        else:
            return ind_x, ind_y


def subset_indices_by_distance(longitude, latitude, centre_lon: float, centre_lat: float, radius: float):
    """
    This method returns a `tuple` of indices within the `radius` of the
    lon/lat point given by the user.
    Scikit-learn BallTree is used to obtain indices.
    :param longitude: The longitude of the users central point
    :param latitude: The latitude of the users central point
    :param radius: The haversine distance (in km) from the central point
    :return: All indices in a `tuple` with the haversine distance of the
            central point
    """

    # Calculate the distances between every model point and the specified
    # centre. Calls another routine dist_haversine.
    dist = calculate_haversine_distance(centre_lon, centre_lat, longitude, latitude)
    indices_bool = dist < radius
    indices = np.where(indices_bool)

    if len(longitude.shape) == 1:
        return xr.DataArray(indices[0])
    else:
        return xr.DataArray(indices[0]), xr.DataArray(indices[1])


def compare_angles(a1, a2, degrees=True):
    """
    # Compares the difference between two angles. e.g. it is 2 degrees between
    # 359 and 1 degree. If degrees = False then will treat angles as radians.
    """

    if not degrees:
        a1 = np.degrees(a1)
        a2 = np.degrees(a2)

    diff = 180 - np.abs(np.abs(a1 - a2) - 180)

    return diff


def cartesian_to_polar(x, y, degrees=True):
    """
    # Conversion of cartesian to polar coordinate system
    # Output theta is in radians
    """
    r = np.sqrt(x**2 + y**2)
    theta = np.arctan2(y, x)
    if degrees:
        theta = np.rad2deg(theta)
    return r, theta


def polar_to_cartesian(r, theta, degrees=True):
    """
    # Conversion of polar to cartesian coordinate system
    # Input theta must be in radians
    """
    if degrees:
        theta = np.deg2rad(theta)
    x = r * np.cos(theta)
    y = r * np.sin(theta)
    return x, y


def subset_indices_lonlat_box(array_lon, array_lat, lon_min, lon_max, lat_min, lat_max):
    ind_lon = np.logical_and(array_lon <= lon_max, array_lon >= lon_min)
    ind_lat = np.logical_and(array_lat <= lat_max, array_lat >= lat_min)
    ind = np.where(np.logical_and(ind_lon, ind_lat))
    return ind


def calculate_haversine_distance(lon1, lat1, lon2, lat2):
    """
    # Estimation of geographical distance using the Haversine function.
    # Input can be single values or 1D arrays of locations. This
    # does NOT create a distance matrix but outputs another 1D array.
    # This works for either location vectors of equal length OR a single loc
    # and an arbitrary length location vector.
    #
    # lon1, lat1 :: Location(s) 1.
    # lon2, lat2 :: Location(s) 2.
    """

    # Convert to radians for calculations
    lon1 = np.deg2rad(lon1)
    lat1 = np.deg2rad(lat1)
    lon2 = np.deg2rad(lon2)
    lat2 = np.deg2rad(lat2)

    # Latitude and longitude differences
    dlat = (lat2 - lat1) / 2
    dlon = (lon2 - lon1) / 2

    # Haversine function.
    distance = np.sin(dlat) ** 2 + np.cos(lat1) * np.cos(lat2) * np.sin(dlon) ** 2
    distance = 2 * 6371.007176 * np.arcsin(np.sqrt(distance))

    return distance


def remove_indices_by_mask(array, mask):
    """
    Removes indices from a 2-dimensional array, A, based on true elements of
    mask. A and mask variable should have the same shape.
    """
    array = np.array(array).flatten()
    mask = np.array(mask, dtype=bool).flatten()
    array_removed = array[~mask]

    return array_removed


def reinstate_indices_by_mask(array_removed, mask, fill_value=np.nan):
    """
    Rebuilds a 2D array from a 1D array created using remove_indices_by_mask().
    False elements of mask will be populated using array_removed. MAsked
    indices will be replaced with fill_value
    """
    array_removed = np.array(array_removed)
    original_shape = mask.shape
    mask = np.array(mask, dtype=bool).flatten()
    array = np.zeros(mask.shape)
    array[~mask] = array_removed
    array[mask] = fill_value
    array = array.reshape(original_shape)
    return array


def nearest_indices_2d(mod_lon, mod_lat, new_lon, new_lat, mask=None):
    """
    Obtains the 2 dimensional indices of the nearest model points to specified
    lists of longitudes and latitudes. Makes use of sklearn.neighbours
    and its BallTree haversine method. Ensure there are no NaNs in
    input longitude/latitude arrays (or mask them using "mask"")

    Example Usage
    ----------
    # Get indices of model points closest to altimetry points
    ind_x, ind_y = nemo.nearest_indices(altimetry.dataset.longitude,
                                        altimetry.dataset.latitude)
    # Nearest neighbour interpolation of model dataset to these points
    interpolated = nemo.dataset.isel(x_dim = ind_x, y_dim = ind_y)

    Parameters
    ----------
    mod_lon (2D array): Model longitude (degrees) array (2-dimensional)
    mod_lat (2D array): Model latitude (degrees) array (2-dimensions)
    new_lon (1D array): Array of longitudes (degrees) to compare with model
    new_lat (1D array): Array of latitudes (degrees) to compare with model
    mask (2D array): Mask array. Where True (or 1), elements of array will
                     not be included. For example, use to mask out land in
                     case it ends up as the nearest point.

    Returns
    -------
    Array of x indices, Array of y indices
    """
    # Cast lat/lon to numpy arrays in case xarray things
    new_lon = np.array(new_lon)
    new_lat = np.array(new_lat)
    mod_lon = np.array(mod_lon)
    mod_lat = np.array(mod_lat)
    original_shape = mod_lon.shape

    # If a mask is supplied, remove indices from arrays.
    if mask is None:
        mod_lon = mod_lon.flatten()
        mod_lat = mod_lat.flatten()
    else:
        mod_lon = remove_indices_by_mask(mod_lon, mask)
        mod_lat = remove_indices_by_mask(mod_lat, mask)
        # If we are masking, we want to preserve the original indices so that
        # we can get them back at the end (since masked points are removed).
        cc, rr = np.meshgrid(np.arange(0, original_shape[1]), np.arange(0, original_shape[0]))
        cc = remove_indices_by_mask(cc, mask)
        rr = remove_indices_by_mask(rr, mask)

    # Put lons and lats into 2D location arrays for BallTree: [lat, lon]
    mod_loc = np.vstack((mod_lat, mod_lon)).transpose()
    new_loc = np.vstack((new_lat, new_lon)).transpose()

    # Convert lat/lon to radians for BallTree
    mod_loc = np.radians(mod_loc)
    new_loc = np.radians(new_loc)

    # Do nearest neighbour interpolation using BallTree (gets indices)
    tree = nb.BallTree(mod_loc, leaf_size=5, metric="haversine")
    _, ind_1d = tree.query(new_loc, k=1)

    if mask is None:
        # Get 2D indices from 1D index output from BallTree
        ind_y, ind_x = np.unravel_index(ind_1d, original_shape)
    else:
        ind_y = rr[ind_1d]
        ind_x = cc[ind_1d]

    ind_x = xr.DataArray(ind_x.squeeze())
    ind_y = xr.DataArray(ind_y.squeeze())

    return ind_x, ind_y


def data_array_time_slice(data_array, date0, date1):
    """Takes an xr.DataArray object and returns a new object with times
    sliced between dates date0 and date1. date0 and date1 may be a string or
    datetime type object."""
    if date0 is None and date1 is None:
        return data_array
    else:
        data_array_sliced = data_array.swap_dims({"t_dim": "time"})
        time_max = data_array.time.max().values
        time_min = data_array.time.min().values
        if date0 is None:
            date0 = time_min
        if date1 is None:
            date1 = time_max
        data_array_sliced = data_array_sliced.sel(time=slice(date0, date1))
        data_array_sliced = data_array_sliced.swap_dims({"time": "t_dim"})
        return data_array_sliced


def day_of_week(date: np.datetime64 = None):
    """Return the day of the week (3 letter str)"""
    if date is None:
        date = np.datetime64("now")

    val = (np.datetime64(date, "D") - np.datetime64(date, "W")).astype(int)
    if val == 0:
        return "Thu"
    elif val == 1:
        return "Fri"
    elif val == 2:
        return "Sat"
    elif val == 3:
        return "Sun"
    elif val == 4:
        return "Mon"
    elif val == 5:
        return "Tue"
    elif val == 6:
        return "Wed"



"""Gridded class"""
import os.path as path_lib
import warnings

# from dask import delayed, compute, visualize
# import graphviz
import gsw
import numpy as np
import xarray as xr

from .._utils import general_utils, stats_util
from .coast import Coast
from .config_parser import ConfigParser
from .._utils.logging_util import get_slug, debug, info, warn, error, warning
import pandas as pd


class Gridded(Coast):  # TODO Complete this docstring
    """
    Words to describe the NEMO class

    kwargs -- define addition keyworded arguemts for domain file. E.g. ln_sco=1
    if using s-scoord in an old domain file that does not carry this flag.
    """

    def __init__(
        self,
        fn_data=None,
        fn_domain=None,  # TODO Super init not called + add a docstring
        multiple=False,
        config: str = " ",
        workers=2,
        threads=2,
        memory_limit_per_worker="2GB",
        **kwargs,
    ):
        debug(f"Creating new {get_slug(self)}")
        self.dataset = xr.Dataset()
        self.grid_ref = None
        self.domain_loaded = False
        self.fn_data = fn_data
        self.fn_domain = fn_domain
        self.grid_vars = None

        if path_lib.isfile(config):
            self.config = ConfigParser(config).config
            if self.config.chunks:
                self._setup_grid_obj(self.config.chunks, multiple, **kwargs)
            else:
                self._setup_grid_obj(None, multiple, **kwargs)
        else:  # allow for usage without config file, this will be limted and dosen't bring the full COAST features
            debug("Config file expected. Limited functionality without config file")
            if self.fn_data is not None:
                self.load(self.fn_data, None, multiple)
            if self.fn_domain is not None:
                self.filename_domain = self.fn_domain
                dataset_domain = self.load_domain(self.fn_domain, None)
                self.dataset["domain"] = dataset_domain

    def _setup_grid_obj(self, chunks, multiple, **kwargs):
        """This is a helper method to reduce the size of def __init__

        Args:
            chunks: This is a setting for xarray as to whether dask (parrell processing) should be on and how it works
            multiple: falg to tell if we are loading one or more files
            **kwargs: pass direct to loaded xarray dataset
        """
        self.set_grid_vars()
        self.set_dimension_mapping()
        self.set_variable_mapping()

        if self.fn_data is not None:
            self.load(self.fn_data, chunks, multiple)

        self.set_dimension_names(self.config.dataset.dimension_map)
        self.set_variable_names(self.config.dataset.variable_map)

        if self.fn_domain is None:
            self.filename_domain = ""  # empty store for domain fileanme
            warn("No NEMO domain specified, only limited functionality" + " will be available")
        else:
            self.filename_domain = self.fn_domain  # store domain fileanme
            dataset_domain = self.load_domain(self.fn_domain, chunks)

            # Define extra domain attributes using kwargs dictionary
            # This is a bit of a placeholder. Some domain/nemo files will have missing variables
            for key, value in kwargs.items():
                dataset_domain[key] = value

            if self.fn_data is not None:
                dataset_domain = self.trim_domain_size(dataset_domain)
            self.set_timezero_depths(
                dataset_domain
            )  # THIS ADDS TO dataset_domain. Should it be 'return'ed (as in trim_domain_size) or is implicit OK?
            self.merge_domain_into_dataset(dataset_domain)
            debug(f"Initialised {get_slug(self)}")

    def set_grid_vars(self):
        """Define the variables to map from the domain file to the NEMO obj"""
        # Define grid specific variables to pull across
        #
        for key, value in self.config.grid_ref.items():
            self.grid_ref = key
            self.grid_vars = value

    # TODO Add parameter type hints and a docstring
    def load_domain(self, fn_domain, chunks):  # TODO Do something with this unused parameter or remove it
        """Loads domain file and renames dimensions with dim_mapping_domain"""
        # Load xarray dataset
        info(f'Loading domain: "{fn_domain}"')
        dataset_domain = xr.open_dataset(fn_domain)
        self.domain_loaded = True
        # Rename dimensions
        for key, value in self.config.domain.dimension_map.items():
            mapping = {key: value}
            try:
                dataset_domain = dataset_domain.rename_dims(mapping)
            except ValueError as err:
                warning(
                    f"{get_slug(self)}: Problem renaming dimension from {get_slug(self.dataset)}: {key} -> {value}."
                    f"{chr(10)}Error message of '{err}'"
                )

        return dataset_domain

    def merge_domain_into_dataset(self, dataset_domain):
        """Merge domain dataset variables into self.dataset, using grid_ref"""
        debug(f"Merging {get_slug(dataset_domain)} into {get_slug(self)}")
        # Define grid independent variables to pull across

        all_vars = self.grid_vars + self.config.code_processing.not_grid_variables

        # Trim domain DataArray area if necessary.
        self.copy_domain_vars_to_dataset(dataset_domain, self.grid_vars)

        # Reset & set specified coordinates
        self.dataset = self.dataset.reset_coords()
        for var in self.config.dataset.coord_var:
            try:
                self.dataset = self.dataset.set_coords(var)
            except ValueError as err:
                warning(f"Issue with settings coordinates using value {var}.{chr(10)}Error message of {err}")

        # Delete specified variables
        for var in self.config.code_processing.delete_variables:
            try:
                self.dataset = self.dataset.drop(var)
            except ValueError as err:
                warning(f"Issue with dropping variable {var}.{chr(10)}Error message of {err}")

    def __getitem__(self, name: str):
        return self.dataset[name]

    def set_grid_ref_attr(self):  # possible not used
        debug(f"{get_slug(self)} grid_ref_attr set to {self.grid_ref_attr_mapping}")
        self.grid_ref_attr_mapping = {
            "temperature": "t-grid",
            "coast_name_for_u_velocity": "u-grid",
            "coast_name_for_v_velocity": "v-grid",
            "coast_name_for_w_velocity": "w-grid",
            "coast_name_for_vorticity": "f-grid",
        }

    def get_contour_complex(self, var, points_x, points_y, points_z, tolerance: int = 0.2):
        debug(f"Fetching contour complex from {get_slug(self)}")
        smaller = self.dataset[var].sel(z=points_z, x=points_x, y=points_y, method="nearest", tolerance=tolerance)
        return smaller

    def set_timezero_depths(self, dataset_domain):
        """
        Calculates the depths at time zero (from the domain_cfg input file)
        for the appropriate grid.
        The depths are assigned to domain_dataset.depth_0
        """
        debug(f"Setting timezero depths for {get_slug(self)} with {get_slug(dataset_domain)}")

        try:
            bathymetry = dataset_domain.bathy_metry.squeeze()
        except AttributeError as err:
            bathymetry = xr.zeros_like(dataset_domain.e1t.squeeze())
            (
                warnings.warn(
                    f"The model domain loaded, '{self.filename_domain}', does not contain the "
                    "bathy_metry' variable. This will result in the "
                    "NEMO.dataset.bathymetry variable being set to zero, which "
                    "may result in unexpected behaviour from routines that require "
                    "this variable."
                )
            )
            debug(
                f"The bathy_metry variable was missing from the domain_cfg for "
                f"{get_slug(self)} with {get_slug(dataset_domain)}"
                f"{chr(10)}Error message of {err}"
            )
        try:
            if self.grid_ref == "t-grid":
                e3w_0 = np.squeeze(dataset_domain.e3w_0.values)
                depth_0 = np.zeros_like(e3w_0)
                depth_0[0, :, :] = 0.5 * e3w_0[0, :, :]
                depth_0[1:, :, :] = depth_0[0, :, :] + np.cumsum(e3w_0[1:, :, :], axis=0)
            elif self.grid_ref == "w-grid":
                e3t_0 = np.squeeze(dataset_domain.e3t_0.values)
                depth_0 = np.zeros_like(e3t_0)
                depth_0[0, :, :] = 0.0
                depth_0[1:, :, :] = np.cumsum(e3t_0, axis=0)[:-1, :, :]
            elif self.grid_ref == "u-grid":
                e3w_0 = dataset_domain.e3w_0.values.squeeze()
                e3w_0_on_u = 0.5 * (e3w_0[:, :, :-1] + e3w_0[:, :, 1:])
                depth_0 = np.zeros_like(e3w_0)
                depth_0[0, :, :-1] = 0.5 * e3w_0_on_u[0, :, :]
                depth_0[1:, :, :-1] = depth_0[0, :, :-1] + np.cumsum(e3w_0_on_u[1:, :, :], axis=0)
                bathymetry[:, :-1] = 0.5 * (bathymetry[:, :-1] + bathymetry[:, 1:])
            elif self.grid_ref == "v-grid":
                e3w_0 = dataset_domain.e3w_0.values.squeeze()
                e3w_0_on_v = 0.5 * (e3w_0[:, :-1, :] + e3w_0[:, 1:, :])
                depth_0 = np.zeros_like(e3w_0)
                depth_0[0, :-1, :] = 0.5 * e3w_0_on_v[0, :, :]
                depth_0[1:, :-1, :] = depth_0[0, :-1, :] + np.cumsum(e3w_0_on_v[1:, :, :], axis=0)
                bathymetry[:-1, :] = 0.5 * (bathymetry[:-1, :] + bathymetry[1:, :])
            elif self.grid_ref == "f-grid":
                e3w_0 = dataset_domain.e3w_0.values.squeeze()
                e3w_0_on_f = 0.25 * (e3w_0[:, :-1, :-1] + e3w_0[:, :-1, 1:] + e3w_0[:, 1:, :-1] + e3w_0[:, 1:, 1:])
                depth_0 = np.zeros_like(e3w_0)
                depth_0[0, :-1, :-1] = 0.5 * e3w_0_on_f[0, :, :]
                depth_0[1:, :-1, :-1] = depth_0[0, :-1, :-1] + np.cumsum(e3w_0_on_f[1:, :, :], axis=0)
                bathymetry[:-1, :-1] = 0.25 * (
                    bathymetry[:-1, :-1] + bathymetry[:-1, 1:] + bathymetry[1:, :-1] + bathymetry[1:, 1:]
                )
            else:
                raise ValueError(str(self) + ": " + self.grid_ref + " depth calculation not implemented")
            # Write the depth_0 variable to the domain_dataset DataSet, with grid type
            dataset_domain[f"depth{self.grid_ref.replace('-grid', '')}_0"] = xr.DataArray(
                depth_0,
                dims=["z_dim", "y_dim", "x_dim"],
                attrs={"units": "m", "standard_name": "Depth at time zero on the {}".format(self.grid_ref)},
            )

            self.dataset["bathymetry"] = bathymetry
            self.dataset["bathymetry"].attrs = {
                "units": "m",
                "standard_name": "bathymetry",
                "description": "depth of last wet w-level on the horizontal {}".format(self.grid_ref),
            }
        except ValueError as err:
            error(err)

    # Add subset method to NEMO class
    def subset_indices(self, *, start: tuple, end: tuple) -> tuple:
        """
        based on transect_indices, this method looks to return all indices between the given points.
        This results in a 'box' (Quadrilateral) of indices.
        consequently the returned lists may have different lengths.
        :param start: A lat/lon pair
        :param end: A lat/lon pair
        :return: list of y indices, list of x indices,
        """
        debug(f"Subsetting {get_slug(self)} indices from {start} to {end}")
        [j1, i1] = self.find_j_i(lat=start[0], lon=start[1])  # lat , lon
        [j2, i2] = self.find_j_i(lat=end[0], lon=end[1])  # lat , lon

        return list(np.arange(j1, j2 + 1)), list(np.arange(i1, i2 + 1))

    def find_j_i(self, *, lat: float, lon: float):
        """
        A routine to find the nearest y x coordinates for a given latitude and longitude
        Usage: [y,x] = find_j_i(lat=49, lon=-12)

        :param lat: latitude
        :param lon: longitude
        :return: the y and x coordinates for the NEMO object's grid_ref, i.e. t,u,v,f,w.
        """
        debug(f"Finding j,i for {lat},{lon} from {get_slug(self)}")
        dist2 = np.square(self.dataset.latitude - lat) + np.square(self.dataset.longitude - lon)
        [y, x] = np.unravel_index(dist2.argmin(), dist2.shape)
        return [y, x]

    def find_j_i_domain(self, *, lat: float, lon: float, dataset_domain: xr.DataArray):
        """
        A routine to find the nearest y x coordinates for a given latitude and longitude
        Usage: [y,x] = find_j_i_domain(lat=49, lon=-12, dataset_domain=dataset_domain)

        :param lat: latitude
        :param lon: longitude
        :param dataset_domain: dataset domain
        :return: the y and x coordinates for the grid_ref variable within the domain file
        """
        debug(f"Finding j,i domain for {lat},{lon} from {get_slug(self)} using {get_slug(dataset_domain)}")
        internal_lat = dataset_domain[self.grid_vars[1]]  # [f"gphi{self.grid_ref.replace('-grid','')}"]
        internal_lon = dataset_domain[self.grid_vars[0]]  # [f"glam{self.grid_ref.replace('-grid','')}"]
        dist2 = np.square(internal_lat - lat) + np.square(internal_lon - lon)
        [_, y, x] = np.unravel_index(dist2.argmin(), dist2.shape)
        return [y, x]

    def transect_indices(self, start: tuple, end: tuple) -> tuple:
        """
        This method returns the indices of a simple straight line transect between two
        lat lon points defined on the NEMO object's grid_ref, i.e. t,u,v,f,w.

        :type start: tuple A lat/lon pair
        :type end: tuple A lat/lon pair
        :return: array of y indices, array of x indices, number of indices in transect
        """
        debug(f"Fetching transect indices for {start} to {end} from {get_slug(self)}")
        [j1, i1] = self.find_j_i(lat=start[0], lon=start[1])  # lat , lon
        [j2, i2] = self.find_j_i(lat=end[0], lon=end[1])  # lat , lon

        line_length = max(np.abs(j2 - j1), np.abs(i2 - i1)) + 1

        jj1 = [int(jj) for jj in np.round(np.linspace(j1, j2, num=line_length))]
        ii1 = [int(ii) for ii in np.round(np.linspace(i1, i2, num=line_length))]

        return jj1, ii1, line_length

    @staticmethod
    def interpolate_in_space(model_array, new_lon, new_lat, mask=None):
        """
        Interpolates a provided xarray.DataArray in space to new longitudes
        and latitudes using a nearest neighbour method (BallTree).

        Example Usage
        ----------
        # Get an interpolated DataArray for temperature onto two locations
        interpolated = nemo.interpolate_in_space(nemo.dataset.votemper,
                                                 [0,1], [45,46])
        Parameters
        ----------
        model_array (xr.DataArray): Model variable DataArray to interpolate
        new_lons (1Darray): Array of longitudes (degrees) to compare with model
        new_lats (1Darray): Array of latitudes (degrees) to compare with model
        mask (2D array): Mask array. Where True (or 1), elements of array will
                     not be included. For example, use to mask out land in
                     case it ends up as the nearest point.

        Returns
        -------
        Interpolated DataArray
        """
        debug(f"Interpolating {get_slug(model_array)} in space with nearest neighbour")
        # Get nearest indices
        ind_x, ind_y = general_utils.nearest_indices_2d(
            model_array.longitude, model_array.latitude, new_lon, new_lat, mask=mask
        )

        # Geographical interpolation (using BallTree indices)
        interpolated = model_array.isel(x_dim=ind_x, y_dim=ind_y)
        if "dim_0" in interpolated.dims:
            interpolated = interpolated.rename({"dim_0": "interp_dim"})
        return interpolated

    @staticmethod
    def interpolate_in_time(model_array, new_times, interp_method="nearest", extrapolate=True):
        """
        Interpolates a provided xarray.DataArray in time to new python
        datetimes using a specified scipy.interpolate method.

        Example Useage
        ----------
        # Get an interpolated DataArray for temperature onto altimetry times
        new_times = altimetry.dataset.time
        interpolated = nemo.interpolate_in_space(nemo.dataset.votemper,
                                                 new_times)
        Parameters
        ----------
        model_array (xr.DataArray): Model variable DataArray to interpolate
        new_times (array): New times to interpolate to (array of datetimes)
        interp_method (str): Interpolation method

        Returns
        -------
        Interpolated DataArray
        """
        debug(f'Interpolating {get_slug(model_array)} in time with method "{interp_method}"')
        # Time interpolation
        interpolated = model_array.swap_dims({"t_dim": "time"})
        if extrapolate:
            interpolated = interpolated.interp(
                time=new_times, method=interp_method, kwargs={"fill_value": "extrapolate"}
            )
        else:
            interpolated = interpolated.interp(time=new_times, method=interp_method)
        # interpolated = interpolated.swap_dims({'time':'t_dim'})  # TODO Do something with this or delete it

        return interpolated

    def construct_density(self, eos="EOS10"):

        """
            Constructs the in-situ density using the salinity, temperture and
            depth_0 fields and adds a density attribute to the t-grid dataset

            Requirements: The supplied t-grid dataset must contain the
            Practical Salinity and the Potential Temperature variables. The depth_0
            field must also be supplied. The GSW package is used to calculate
            The Absolute Pressure, Absolute Salinity and Conservate Temperature.

            Note that currently density can only be constructed using the EOS10
            equation of state.

        Parameters
        ----------
        eos : equation of state, optional
            DESCRIPTION. The default is 'EOS10'.


        Returns
        -------
        None.
        adds attribute NEMO.dataset.density

        """
        debug(f'Constructing in-situ density for {get_slug(self)} with EOS "{eos}"')
        try:
            if eos != "EOS10":
                raise ValueError(str(self) + ": Density calculation for " + eos + " not implemented.")
            if self.grid_ref != "t-grid":
                raise ValueError(
                    str(self)
                    + ": Density calculation can only be performed for a t-grid object,\
                                 the tracer grid for NEMO."
                )

            try:
                shape_ds = (
                    self.dataset.t_dim.size,
                    self.dataset.z_dim.size,
                    self.dataset.y_dim.size,
                    self.dataset.x_dim.size,
                )
                sal = self.dataset.salinity.to_masked_array()
                temp = self.dataset.temperature.to_masked_array()
            except AttributeError:
                shape_ds = (1, self.dataset.z_dim.size, self.dataset.y_dim.size, self.dataset.x_dim.size)
                sal = self.dataset.salinity.to_masked_array()[np.newaxis, ...]
                temp = self.dataset.temperature.to_masked_array()[np.newaxis, ...]

            density = np.ma.zeros(shape_ds)

            s_levels = self.dataset.depth_0.to_masked_array()
            lat = self.dataset.latitude.values
            lon = self.dataset.longitude.values
            # Absolute Pressure
            pressure_absolute = np.ma.masked_invalid(gsw.p_from_z(-s_levels, lat))  # depth must be negative
            # Absolute Salinity
            sal_absolute = np.ma.masked_invalid(gsw.SA_from_SP(sal, pressure_absolute, lon, lat))
            sal_absolute = np.ma.masked_less(sal_absolute, 0)
            # Conservative Temperature
            temp_conservative = np.ma.masked_invalid(gsw.CT_from_pt(sal_absolute, temp))
            # In-situ density
            density = np.ma.masked_invalid(gsw.rho(sal_absolute, temp_conservative, pressure_absolute))

            coords = {
                "depth_0": (("z_dim", "y_dim", "x_dim"), self.dataset.depth_0.values),
                "latitude": (("y_dim", "x_dim"), self.dataset.latitude.values),
                "longitude": (("y_dim", "x_dim"), self.dataset.longitude.values),
            }
            dims = ["z_dim", "y_dim", "x_dim"]
            attributes = {"units": "kg / m^3", "standard name": "In-situ density"}

            if shape_ds[0] != 1:
                coords["time"] = (("t_dim"), self.dataset.time.values)
                dims.insert(0, "t_dim")

            self.dataset["density"] = xr.DataArray(np.squeeze(density), coords=coords, dims=dims, attrs=attributes)

        except AttributeError as err:
            error(err)

    def trim_domain_size(self, dataset_domain):
        """
        Trim the domain variables if the dataset object is a spatial subset

        Note: This breaks if the SW & NW corner values of nav_lat and nav_lon
        are masked, as can happen if on land...
        """
        debug(f"Trimming {get_slug(self)} variables with {get_slug(dataset_domain)}")
        if (self.dataset["x_dim"].size != dataset_domain["x_dim"].size) or (
            self.dataset["y_dim"].size != dataset_domain["y_dim"].size
        ):
            info(
                "The domain  and dataset objects are different sizes:"
                " [{},{}] cf [{},{}]. Trim domain.".format(
                    dataset_domain["x_dim"].size,
                    dataset_domain["y_dim"].size,
                    self.dataset["x_dim"].size,
                    self.dataset["y_dim"].size,
                )
            )

            # Find the corners of the cut out domain.
            [j0, i0] = self.find_j_i_domain(
                lat=self.dataset.nav_lat[0, 0], lon=self.dataset.nav_lon[0, 0], dataset_domain=dataset_domain
            )
            [j1, i1] = self.find_j_i_domain(
                lat=self.dataset.nav_lat[-1, -1], lon=self.dataset.nav_lon[-1, -1], dataset_domain=dataset_domain
            )

            dataset_subdomain = dataset_domain.isel(y_dim=slice(j0, j1 + 1), x_dim=slice(i0, i1 + 1))
            return dataset_subdomain
        else:
            return dataset_domain

    def copy_domain_vars_to_dataset(self, dataset_domain, grid_vars):
        """
        Map the domain coordand metric variables to the dataset object.
        Expects the source and target DataArrays to be same sizes.
        """
        debug(f"Copying domain vars from {get_slug(dataset_domain)}/{get_slug(grid_vars)} to {get_slug(self)}")
        for var in grid_vars:
            try:
                new_name = self.config.domain.variable_map[var]
                self.dataset[new_name] = dataset_domain[var].squeeze()
                debug("map: {} --> {}".format(var, new_name))
            except:  # FIXME Catch specific exception(s)
                pass  # TODO Should we log something here?

    def differentiate(self, in_var_str, config_path=None, dim="z_dim", out_var_str=None, out_obj=None):
        """
        Derivatives are computed in x_dim, y_dim, z_dim (or i,j,k) directions
        wrt lambda, phi, or z coordinates (with scale factor in metres not degrees).

        Derivatives are calculated using the approach adopted in NEMO,
        specifically using the 1st order accurate central difference
        approximation. For reference see section 3.1.2 (sec. Discrete operators)
        of the NEMO v4 Handbook.

        Currently the method does not accomodate all possible eventualities. It
        covers:
        1) d(grid_t)/dz --> grid_w

        Returns  an object (with the appropriate target grid_ref) containing
        derivative (out_var_str) as xr.DataArray

        This is hardwired to expect:
        1) depth_0 and e3_0 fields exist
        2) xr.DataArrays are 4D
        3) self.filename_domain if out_obj not specified
        4) If out_obj is not specified, one is built that is  the size of
            self.filename_domain. I.e. automatic subsetting of out_obj is not
            supported.

        Example usage:
        --------------
        # Initialise DataArrays
        nemo_t = coast.NEMO( fn_data, fn_domain, grid_ref='t-grid' )
        # Compute dT/dz
        nemo_w_1 = nemo_t.differentiate( 'temperature', dim='z_dim' )

        # For f(z)=-z. Compute df/dz = -1. Surface value is set to zero
        nemo_t.dataset['depth4D'],_ = xr.broadcast( nemo_t.dataset['depth_0'], nemo_t.dataset['temperature'] )
        nemo_w_4 = nemo_t.differentiate( 'depth4D', dim='z_dim', out_var_str='dzdz' )

        Provide an existing target NEMO object and target variable name:
        nemo_w_1 = nemo_t.differentiate( 'temperature', dim='z_dim', out_var_str='dTdz', out_obj=nemo_w_1 )


        Parameters
        ----------
        in_var_str : str, name of variable to differentiate
        config_path : str, path to the w grid config file
        dim : str, dimension to operate over. E.g. {'z_dim', 'y_dim', 'x_dim', 't_dim'}
        out_var_str : str, (optional) name of the target xr.DataArray
        out_obj : exiting NEMO obj to store xr.DataArray (optional)

        """
        import xarray as xr

        new_units = ""

        # Check in_var_str exists in self.
        if hasattr(self.dataset, in_var_str):
            # self.dataset[in_var_str] exists

            var = self.dataset[in_var_str]  # for convenience

            nt = var.sizes["t_dim"]
            nz = var.sizes["z_dim"]
            ny = var.sizes["y_dim"]
            nx = var.sizes["x_dim"]

            # Compute d(t_grid)/dz --> w-grid
            # Check grid_ref and dir. Determine target grid_ref.
            if (self.grid_ref == "t-grid") and (dim == "z_dim"):
                out_grid = "w-grid"

                # If out_obj exists check grid_ref, else create out_obj.
                if (out_obj is None) or (out_obj.grid_ref != out_grid):
                    try:
                        out_obj = Gridded(fn_domain=self.filename_domain, config=config_path)
                    except:  # TODO Catch specific exception(s)
                        warn(
                            "Failed to create target NEMO obj. Perhaps self.",
                            "filename_domain={} is empty?".format(self.filename_domain),
                        )

                # Check is out_var_str is defined, else create it
                if out_var_str is None:
                    out_var_str = in_var_str + "_dz"

                # Create new DataArray with the same dimensions as the parent
                # Crucially have a coordinate value that is appropriate to the target location.
                blank = xr.zeros_like(var.isel(z_dim=[0]))  # Using "z_dim=[0]" as a list preserves z-dimension
                blank.coords["depth_0"] -= blank.coords["depth_0"]  # reset coord vals to zero
                # Add blank slice to the 'surface'. Concat over the 'dim' coords
                diff = xr.concat([blank, var.diff(dim)], dim)
                diff_ndim, e3w_ndim = xr.broadcast(diff, out_obj.dataset.e3_0.squeeze())
                # Compute the derivative
                out_obj.dataset[out_var_str] = -diff_ndim / e3w_ndim

                # Assign attributes
                new_units = var.units + "/" + out_obj.dataset.depth_0.units
                # Convert to a xr.DataArray and return
                out_obj.dataset[out_var_str].attrs = {"units": new_units, "standard_name": out_var_str}

                # Return in object.
                return out_obj

            else:
                warn("Not ready for that combination of grid ({}) and " "derivative ({})".format(self.grid_ref, dim))
                return None
        else:
            warn(f"{in_var_str} does not exist in {get_slug(self)} dataset")
            return None

    def apply_doodson_x0_filter(self, var_str):
        """Applies Doodson X0 filter to a variable.

        Input variable is expected to be hourly.
        Output is saved back to original dataset as {var_str}_dxo

        !!WARNING: Will load in entire variable to memory. If dataset large,
        then subset before using this method or ensure you have enough free
        RAM to hold the variable (twice).

        DB:: Currently not tested in unit_test.py"""
        var = self.dataset[var_str]
        new_var_str = var_str + "_dx0"
        old_dims = var.dims
        time_index = old_dims.index("t_dim")
        filtered = stats_util.doodson_x0_filter(var, ax=time_index)
        if filtered is not None:
            self.dataset[new_var_str] = (old_dims, filtered)
        return

    @staticmethod
    def get_e3_from_ssh(nemo_t, e3t=True, e3u=False, e3v=False, e3f=False, e3w=False, dom_fn: str = None):
        """
        Where the model has been run with a nonlinear free surface
        and z* variable volumne (ln_vvl_zstar=True) then the vertical scale factors
        will vary in time (and space). This function will compute the vertical
        scale factors e3t, e3u, e3v, e3f and e3w by using the sea surface height
        field (ssh variable) and initial scale factors from the domain_cfg file.
        The vertical scale factors will be computed at the same model time as the
        ssh and if the ssh field is averaged in time then the scale factors will
        also be time averages.

        A t-grid NEMO object containing the ssh variable must be passed in. Either
        the domain_cfg path must have been passed in as an argument when the NEMO
        object was created or it must be passed in here using the dom_fn argument.

        e.g. e3t,e3v,e3f = coast.NEMO.get_e3_from_ssh(nemo_t,true,false,true,true,false)

        Parameters
        ----------
        nemo_t : (Coast.NEMO), NEMO object on the t-grid containing the ssh variable
        e3t : (boolean), true if e3t is to be returned. Default True.
        e3u : (boolean), true if e3u is to be returned. Default False.
        e3v : (boolean), true if e3v is to be returned. Default False.
        e3f : (boolean), true if e3f is to be returned. Default False.
        e3w : (boolean), true if e3w is to be returned. Default False.
        dom_fn : (str), Optional, path to domain_cfg file.

        Returns
        -------
        Tuple of xarray.DataArrays
        (e3t, e3u, e3v, e3f, e3w)
        Only those requested will be returned, but the ordering is always the same.

        """
        e3_return = []
        try:
            ssh = nemo_t.dataset.ssh
        except AttributeError:
            print("The nemo_t dataset must contain the ssh variable.")
            return
        if "t_dim" not in ssh.dims:
            ssh = ssh.expand_dims("t_dim", axis=0)

        # Load domain_cfg
        if dom_fn is None:
            dom_fn = nemo_t.filename_domain
        try:
            ds_dom = xr.open_dataset(dom_fn).squeeze().rename({"z": "z_dim", "x": "x_dim", "y": "y_dim"})
        except OSError:
            print(f"Problem opening domain_cfg file: {dom_fn}")
            return

        e3t_0 = ds_dom.e3t_0

        # Water column thickness, i.e. depth of bottom w-level on horizontal t-grid
        H = e3t_0.cumsum(dim="z_dim").isel(z_dim=ds_dom.bottom_level.astype("int") - 1)
        # Add correction to e3t_0 due to change in ssh
        e3t_new = e3t_0 * (1 + ssh / H)
        # preserve dimension ordering
        e3t_new = e3t_new.transpose("t_dim", "z_dim", "y_dim", "x_dim")
        # mask out correction at layers below bottom level
        e3t_new = e3t_new.where(e3t_new.z_dim < ds_dom.bottom_level, e3t_0.data)
        # preserve any other t mask
        e3t_new = e3t_new.where(~np.isnan(ssh))
        if e3t:
            e3_return.append(e3t_new.squeeze())

        if np.any([e3u, e3v, e3f]):
            e1e2t = ds_dom.e1t * ds_dom.e2t
        if np.any([e3u, e3v, e3w]):
            e3t_dt = e3t_new - e3t_0

        # area averaged interpolation onto the u-grid to get e3u
        if np.any([e3u, e3f]):
            e1e2u = ds_dom.e1u * ds_dom.e2u
            # interpolate onto u-grid
            e3u_temp = (
                (0.5 / e1e2u[:, :-1]) * ((e1e2t[:, :-1] * e3t_dt[:, :, :, :-1]) + (e1e2t[:, 1:] * e3t_dt[:, :, :, 1:]))
            ).transpose("t_dim", "z_dim", "y_dim", "x_dim")
            # u mask
            e3u_temp = e3u_temp.where(e3t_dt[:, :, :, 1:] != 0, 0)
            # mask out correction at layers below bottom level
            e3u_temp = e3u_temp.where(e3u_temp.z_dim < ds_dom.bottom_level[:, :-1], 0)
            # Add correction to e3u_0
            e3u_temp = e3u_temp + ds_dom.e3u_0[:, :, :-1]
            e3u_new = xr.zeros_like(e3t_new)
            e3u_new = e3u_new.load()
            e3u_new[:, :, :, :-1] = e3u_temp
            e3u_new[:, :, :, -1] = ds_dom.e3u_0[:, :, -1]
            e3u_new["longitude"] = ds_dom.glamu
            e3u_new["latitude"] = ds_dom.gphiu
            if e3u:
                e3_return.append(e3u_new.squeeze())

        # area averaged interpolation onto the u-grid to get e3v
        if e3v:
            e1e2v = ds_dom.e1v * ds_dom.e2v
            e3v_temp = (
                (0.5 / e1e2v[:-1, :]) * ((e1e2t[:-1, :] * e3t_dt[:, :, :-1, :]) + (e1e2t[1:, :] * e3t_dt[:, :, 1:, :]))
            ).transpose("t_dim", "z_dim", "y_dim", "x_dim")
            e3v_temp = e3v_temp.where(e3t_dt[:, :, 1:, :] != 0, 0)
            e3v_temp = e3v_temp.where(e3v_temp.z_dim < ds_dom.bottom_level[:-1, :], 0)
            e3v_temp = e3v_temp + ds_dom.e3v_0[:, :-1, :]
            e3v_new = xr.zeros_like(e3t_new)
            e3v_new = e3v_new.load()
            e3v_new[:, :, :-1, :] = e3v_temp
            e3v_new[:, :, -1, :] = ds_dom.e3v_0[:, -1, :]
            e3v_new["longitude"] = ds_dom.glamv
            e3v_new["latitude"] = ds_dom.gphiv
            e3_return.append(e3v_new.squeeze())

        # area averaged interpolation onto the u-grid to get e3f
        if e3f:
            e1e2f = ds_dom.e1f * ds_dom.e2f
            e3u_dt = e3u_new - ds_dom.e3u_0
            e3f_temp = (
                (0.5 / e1e2f[:-1, :]) * ((e1e2u[:-1, :] * e3u_dt[:, :, :-1, :]) + (e1e2u[1:, :] * e3u_dt[:, :, 1:, :]))
            ).transpose("t_dim", "z_dim", "y_dim", "x_dim")
            e3f_temp = e3f_temp.where(e3u_dt[:, :, 1:, :] != 0, 0)
            e3f_temp = e3f_temp.where(e3f_temp.z_dim < ds_dom.bottom_level[:-1, :], 0)
            e3f_temp = e3f_temp + ds_dom.e3f_0[:, :-1, :]
            e3f_new = xr.zeros_like(e3t_new)
            e3f_new = e3f_new.load()
            e3f_new[:, :, :-1, :] = e3f_temp
            e3f_new[:, :, -1, :] = ds_dom.e3f_0[:, -1, :]
            e3f_new["longitude"] = ds_dom.glamf
            e3f_new["latitude"] = ds_dom.gphif
            e3_return.append(e3f_new.squeeze())

        # simple vertical interpolation for e3w. Special treatment of top and bottom levels
        if e3w:
            # top levels correction same at e3t
            e3w_new = (ds_dom.e3w_0 + e3t_dt).transpose("t_dim", "z_dim", "y_dim", "x_dim")
            # levels between top and bottom
            e3w_new = e3w_new.load()
            e3w_new[dict(z_dim=slice(1, None))] = (
                0.5 * e3t_dt[:, :-1, :, :] + 0.5 * e3t_dt[:, 1:, :, :] + ds_dom.e3w_0[1:, :, :]
            )
            # bottom and below levels
            e3w_new = e3w_new.where(e3w_new.z_dim < ds_dom.bottom_level, e3t_dt.shift(z_dim=1) + ds_dom.e3w_0)
            e3_return.append(e3w_new.squeeze())

        return tuple(e3_return)

    def harmonics_combine(self, constituents, components=["x", "y"]):
        """
        Contains a new NEMO object containing combined harmonic information
        from the original object.

        NEMO saves harmonics to individual variables such as M2x, M2y... etc.
        This routine will combine these variables (depending on constituents)
        into a single data array. This new array will have the new dimension
        'constituent' and a new data coordinate 'constituent_name'.

        Parameters
        ----------
        constituents : List of strings containing constituent names to combine.
                       The case of these strings should match that used in
                       NEMO output. If a constituent is not found, no problem,
                       it just won't be in the combined dataset.
        components   : List of strings containing harmonic components to look
                       for. By default, this looks for the complex components
                       'x' and 'y'. E.g. if constituents = ['M2'] and
                       components is left as default, then the routine looks
                       for ['M2x', and 'M2y'].

        Returns
        -------
        NEMO() object, containing combined harmonic variables in a new dataset.
        """

        # Select only the specified constituents. NEMO model harmonics names are
        # things like "M2x" and "M2y". Ignore current harmonics. Start by constructing
        # the possible variable names
        names_x = np.array([cc + components[0] for cc in constituents])
        names_y = np.array([cc + components[1] for cc in constituents])
        constituents = np.array(constituents, dtype="str")

        # Compare against names in file
        var_keys = np.array(list(self.dataset.keys()))
        indices = [np.where(names_x == ss) for ss in names_x if ss in var_keys]
        indices = np.array(indices).T.squeeze()

        # Index the possible names to match file names
        names_x = names_x[indices]
        names_y = names_y[indices]
        constituents = constituents[indices]

        # Concatenate x and y variables into one array
        x_arrays = [self.dataset[ss] for ss in names_x]
        harmonic_x = "harmonic_" + components[0]
        x_data = xr.concat(x_arrays, dim="constituent").rename(harmonic_x)
        y_arrays = [self.dataset[ss] for ss in names_y]
        harmonic_y = "harmonic_" + components[1]
        y_data = xr.concat(y_arrays, dim="constituent").rename(harmonic_y)

        nemo_harmonics = Gridded()
        nemo_harmonics.dataset = xr.merge([x_data, y_data])
        nemo_harmonics.dataset["constituent"] = constituents

        return nemo_harmonics

    def harmonics_convert(
        self,
        direction="cart2polar",
        x_var="harmonic_x",
        y_var="harmonic_y",
        a_var="harmonic_a",
        g_var="harmonic_g",
        degrees=True,
    ):
        """
        Converts NEMO harmonics from cartesian to polar or vice versa.
        Make sure this NEMO object contains combined harmonic variables
        obtained using harmonics_combine().

        *Note:

        Parameters
        ----------
        direction (str) : Choose 'cart2polar' or 'polar2cart'. If 'cart2polar'
                          Then will look for variables x_var and y_var. If
                          polar2cart, will look for a_var (amplitude) and
                          g_var (phase).
        x_var (str)     : Harmonic x variable name in dataset (or output)
                          default = 'harmonic_x'.
        y_var (str)     : Harmonic y variable name in dataset (or output)
                          default = 'harmonic_y'.
        a_var (str)     : Harmonic amplitude variable name in dataset (or output)
                          default = 'harmonic_a'.
        g_var (str)     : Harmonic phase variable name in dataset (or output)
                          default = 'harmonic_g'.
        degrees (bool)  : Whether input/output phase are/will be in degrees.
                          Default is True.

        Returns
        -------
        Modifies NEMO() dataset in place. New variables added.
        """
        if direction == "cart2polar":
            a, g = general_utils.cartesian_to_polar(self.dataset[x_var], self.dataset[y_var], degrees=degrees)
            self.dataset[a_var] = a
            self.dataset[g_var] = g
        elif direction == "polar2cart":
            x, y = general_utils.polar_to_cartesian(self.dataset[a_var], self.dataset[g_var], degrees=degrees)
            self.dataset[x_var] = x
            self.dataset[y_var] = y
        else:
            print("Unknown direction setting. Choose cart2polar or polar2cart")

        return

    def time_slice(self, date0, date1):
        """Return new Gridded object, indexed between dates date0 and date1"""
        dataset = self.dataset
        t_ind = pd.to_datetime(dataset.time.values) >= date0
        dataset = dataset.isel(t_dim=t_ind)
        t_ind = pd.to_datetime(dataset.time.values) < date1
        dataset = dataset.isel(t_dim=t_ind)
        gridded_out = Gridded()
        gridded_out.dataset = dataset
        return gridded_out
